# -*- coding: utf-8 -*-
"""
@author: Erwan

Logging and warning functions
"""

from __future__ import absolute_import, division, print_function, unicode_literals

from time import gmtime, strftime
from warnings import warn
from termcolor import colored
import sys


def printlog(msg, logfile='log.txt', stdout=True):
    ''' Write a message to the logfile, adding date and time. Also print it to screen
    if stdout is True '''
    with open(logfile, 'a+') as f:
        f.write('{0} {1}\n'.format(
            strftime("%Y-%m-%d %H:%M:%S", gmtime()), msg))
    if stdout:
        print(msg)
    return


def warnlog(msg, logfile='log.txt', stdout=True):
    ''' Write a WARNING message to the logfile, adding date and time. Also print it to 
    screen if stdout is True '''
    with open(logfile, 'a+') as f:
        f.write('{0} WARNING. {1}\n'.format(
            strftime("%Y-%m-%d %H:%M:%S", gmtime()), msg))
    if stdout:
        warn(msg)
    return


def printwarn(msg, verbose=True, warnings=True):
    ''' Send msg in the console if verbose is True, and as warning if warnings 
    is True. Warnings is added at the end of the process but you may miss it 
    if interrupted. 
    '''

    if verbose:
        print(colored('\n'+msg+'\n', 'red'))
        sys.stdout.flush()
    if warnings:
        warn(msg)
    return



# -*- coding: utf-8 -*-
"""
Test names and labels
"""

from __future__ import absolute_import, unicode_literals, division, print_function
from radis.lbl.labels import (vib_lvl_name_hitran_class1, vib_lvl_name_hitran_class5,
                             vib_lvl_name_hitran_class5_short, vib_lvl_name_cdsd_pc,
                             vib_lvl_name_cdsd_pcN)
from radis.misc.printer import printm
import pytest


@pytest.mark.fast
def test_vibrational_levels_labelling(verbose=True, *args, **kwargs):

    # If verbose: print some
    if verbose:
        printm('Some vibrational level formats:')

        printm('... CO, HITRAN format (v):\t\t',
               vib_lvl_name_hitran_class1(10))

        printm('... CO2, HITRAN format (v1,v2,l2,v3):\t\t',
               vib_lvl_name_hitran_class5(2, 1, 1, 3))
        printm('... CO2, HITRAN format, short (v1v2l2v3):\t',
               vib_lvl_name_hitran_class5_short(2, 1, 1, 3))

        printm('... CO2, CDSD format (p,c):\t\t', vib_lvl_name_cdsd_pc(14, 1))
        printm('... CO2, CDSD format (p,c,N):\t\t',
               vib_lvl_name_cdsd_pcN(14, 1, 1))

    # Do the tests

    assert vib_lvl_name_hitran_class1(10) == '(10)'

    assert vib_lvl_name_hitran_class5(2, 1, 1, 3) == '(2,1,1,3)'
    assert vib_lvl_name_hitran_class5_short(2, 1, 1, 3) == '21`1`3'

    assert vib_lvl_name_cdsd_pc(14, 1) == '(14,1)'
    assert vib_lvl_name_cdsd_pcN(14, 1, 1) == '(14,1,1)'

    return


if __name__ == '__main__':

    test_vibrational_levels_labelling()



# -*- coding: utf-8 -*-
"""
Created on Tue Jul 31 16:28:43 2018

@author: erwan
"""


def morse_increment(Dzero, we, wexe=0, weye=0, weze=0,
                       weae=0, webe=0, wece=0):
    ''' Get Morse potential energy increment correction for given molecule

    Parameters
    ----------
    
    Dzero: cm-1
        dissociation energy
        
    we, wexe, etc.: cm-1
        spectroscopic parameters

    Notes
    -----
    
    Internal:
        
        Dzero = self.Ediss
        d_equil = (Dzero+0.5*we-0.25*wexe+0.125*weye
                   + 0.0625*weze+0.03125*weae+0.015625*webe
                   + 7.8125e-3*wece)

        v_inc = we**2 / (2*d_equil)

    References
    -----------

    Specair
    
    '''

    d_equil = (Dzero+0.5*we-0.25*wexe+0.125*weye
               + 0.0625*weze+0.03125*weae+0.015625*webe
               + 7.8125e-3*wece)

    v_inc = we**2 / (2*d_equil)
    
    return v_inc



# -*- coding: utf-8 -*-
"""
Created on Sat Oct 28 18:47:39 2017

@author: erwan

State-dependant and state-independant degeneracies for molecules 


"""
#TODO: Make it a JSON file

from __future__ import print_function, absolute_import, division, unicode_literals

def gi(M, I):
    ''' State independant degeneracy 

    Parameters
    ----------
    
    M: int
        molecule id
        
    I: int
        isotope number

    References
    ----------

    Šimečková 2006: "Einstein A-coefficients and statistical weights for molecular
    absorption transitions in the HITRAN database". DOI 10.1016/j.jqsrt.2005.07.003

    '''

    _gi = {
        2:  # CO2
            {1: 1,   # 626
             2: 2,   # 636
             3: 1,   # 628
             4: 6,   # 627
         },
        5:   # CO
            {1: 1,   # 26
             2: 2,   # 36
             3: 1,
             4: 6,
         },
    }

    try:
        return _gi[M][I]
    except KeyError:
        raise NotImplementedError('undefined state-independant degeneracy for ' +
                                  'molecule[isotope]: {0}[{1}]'.format(M, I))


def gs(M, I):
    ''' State dependant degeneracy

    Parameters
    ----------
    
    M: int
        molecule id
        
    I: int
        isotope number

    References
    ----------

    Šimečková 2006: "Einstein A-coefficients and statistical weights for molecular
    absorption transitions in the HITRAN database". DOI 10.1016/j.jqsrt.2005.07.003

    CO2::

        For the 12C16O2, 13C16O2, 12C18O2 isotopologues, the statistical weights gs
        of the symmetric s and antisymmetric a rotational levels are 1 and 0, respectively
        (see Fig. 99 of Ref. [24], Fig. 17-6 of Ref. [26]). For the 16O12C18O, 16O12C17O,
        16O13C18O, 16O13C17O, 17O12C18O isotopologues, the statistical weights gs
        of all the rotational levels are 1"

    '''

    _gs = {
        2:  # CO2
            {1: (1,0),   # 626       (normally 1:0, but negative levels dont exist)
             2: (1,0),   # 636       (normally 1:0, but negative levels dont exist)
             3: 1,       # 628 
             4: 1,       # 627
             },
        5:   # CO
            {1: 1,   # 26
             2: 1,   # 36
             3: 1,
             4: 1,
             },
    }

    try:
        return _gs[M][I]
    except KeyError:
        raise NotImplementedError('undefined state-dependant degeneracy for ' +
                                  'molecule[isotope]: {0}[{1}]'.format(M, I))




# -*- coding: utf-8 -*-
"""
Created on Tue Nov 21 08:56:12 2017

@author: erwan

Use
------

Run all tests:

>>> pytest       (in command line, in project folder)

Run only fast tests (i.e: tests that have a 'fast' label)

>>> pytest -m fast


----------


"""

from __future__ import absolute_import, unicode_literals, division, print_function
from radis.lbl.overp import LevelsList
from radis.lbl import SpectrumFactory
from radis.misc.printer import printm
from radis.spectrum import plot_diff
from radis.misc import DatabankNotFound
from radis.test.utils import IgnoreMissingDatabase, setup_test_line_databases
import numpy as np
import matplotlib.pyplot as plt
import pytest

@pytest.mark.needs_config_file
@pytest.mark.needs_db_CDSD_HITEMP_PCN
#@pytest.mark.needs_connection
def test_direct_overpopulation_vs_recombined_bands(verbose=True, plot=False, warnings=True,
                                                   rtol=0.05, *args, **kwargs):
    ''' Compare a non-equilibrium spectrum calculated directly with overpopulations, 
    or by recombining pre-calculated vibrational bands. 

    The later allows for almost instantaneous changes of the overpopulation factors, 
    (mostly useful in fitting algorithms), but is only valid for optically thin emission spectra

    Expected output:

    when x_CO2 = 1e-3, radiance in both cases match
    when x_CO2 = 1, they dont

    '''

#    Notes
#    -----
#    
#    On the old NeQ package the test used [HITEMP-2010]_
#    
#    Starting from RADIS 1.0.1, the test is run on [HITRAN-2016]_, which 
#    is not valid for these temperatures but can be more conveniently 
#    downloaded automatically and thus executed everytime with [Travis]_
#    
    
    # Note: only with iso1 at the moment

    if plot:  # Make sure matplotlib is interactive so that test are not stuck in pytest
        plt.ion()

    try:

        # Generate factory
        iso = 1
        sf = SpectrumFactory(
            wavelength_min=4220,
            wavelength_max=4280,
            mole_fraction=1e-3,
            path_length=10,
            cutoff=1e-25,
            molecule='CO2',
            isotope=iso,
            db_use_cached=True,
            wstep=0.01,
            broadening_max_width=5,
            medium='air',
            verbose=verbose,
        )
        sf.warnings['MissingSelfBroadeningWarning'] = 'ignore'
        sf.warnings['NegativeEnergiesWarning'] = 'ignore'
        sf.load_databank('CDSD-HITEMP-PCN')
#        sf.fetch_databank()   # uses HITRAN: not really valid at this temperature, but runs on all machines without install

        # Generate bands to recombine
        parfunc = sf.parsum_calc['CO2'][iso]['X']
        Tref = 1500
        # , return_lines=False)
        s_bands = sf.non_eq_bands(Tvib=Tref, Trot=Tref)
        lvlist = LevelsList(parfunc, s_bands, sf.params.levelsfmt)

        # Compare ab initio and recombined from bands at M = 1e-3
        s_recombined = lvlist.non_eq_spectrum(
            Tvib=Tref, Trot=Tref, overpopulation={'(4,1,3)': 3})
        sref = sf.non_eq_spectrum(
            Tvib=Tref, Trot=Tref, overpopulation={'(4,1,3)': 1})

        if verbose:
            printm('Testing x_CO2 = 1e-3: ab initio ~ recombined bands (<{0:.1f}%):\t'.format(
                rtol*100))

        if plot:
            plot_diff(sref, s_recombined, var='radiance_noslit', label1='ab initio', label2='recombined bands',
                      title='x_CO2 = 1e-3')

        assert np.allclose(s_recombined.get_radiance_noslit(), sref.get_radiance_noslit(),
                           rtol=rtol)

        # Rescale and try again for x_CO2 = 1
        s_recombined.rescale_mole_fraction(1)
        sref.rescale_mole_fraction(1)

        if plot:
            plot_diff(sref, s_recombined, var='radiance_noslit', label1='ab initio', label2='recombined bands',
                      title='x_CO2 = 1')

        if verbose:
            printm('Testing x_CO2 = 1: ab initio ~ recombined bands (<{0:.1f}%):\t{1}'.format(
                rtol*100, np.allclose(s_recombined.get_radiance_noslit(), sref.get_radiance_noslit(),
                                      rtol=rtol)))

        with pytest.raises(AssertionError):
            assert np.allclose(s_recombined.get_radiance_noslit(), sref.get_radiance_noslit(),
                               rtol=rtol)

        return True

    except DatabankNotFound as err:
        assert IgnoreMissingDatabase(err, __file__, warnings)


def test_3Tvib_vs_1Tvib(verbose=True, plot=False, warnings=True, *args, **kwargs):

    if plot:  # Make sure matplotlib is interactive so that test are not stuck in pytest
        plt.ion()
        
    setup_test_line_databases()

    try:
        T = 1500

        iso = [1]  
        sf = SpectrumFactory(
            wavenum_min=2380, 
            wavenum_max=2400,
            pressure=20*1e-3,
            db_use_cached=True,
            cutoff=1e-25,
            isotope=iso,  # ,2',
            path_length=10,
            mole_fraction=0.1,
            broadening_max_width=1,
            medium='vacuum',
            verbose=verbose,
        )
        sf.warnings.update({'MissingSelfBroadeningWarning': 'ignore',
                            'VoigtBroadeningWarning': 'ignore'})
        sf.load_databank('HITRAN-CO2-TEST')

        # Compare energies
        for I in iso:
            energies = sf.get_energy_levels('CO2', I, 'X')
            assert (energies.Evib == energies.Evib1 +
                    energies.Evib2+energies.Evib3).all()
        if verbose:
            printm('Tested Evib == Evib1+Evib2+Evib3: OK')

        s3 = sf.non_eq_spectrum((T, T, T), T)
        s1 = sf.non_eq_spectrum(T, T)
        s3.name = 'Tvib=({0},{0},{0}) K, Trot={0} K'.format(T)
        s1.name = 'Tvib={0} K, Trot={0} K'.format(T)

        if plot:
            s1.plot('transmittance_noslit', wunit='cm-1', color='k', lw=3,
                    label='Tvib = {0}K'.format(T))

            s3.plot('transmittance_noslit', wunit='cm-1', color='r', lw=3,
                    nfig='same', label='Tvib1 = Tvib2 = Tvib3 = {0}K'.format(T))

        assert s1.compare_with(s3, spectra_only=True,
                               verbose=verbose, plot=plot)
        if verbose:
            printm('Tested Spectra 3Tvib(T1=T2=T3=T) and 1Tvib(T) are the same: OK')

        return True

    except DatabankNotFound as err:
        assert IgnoreMissingDatabase(err, __file__, warnings)


def run_testcases(verbose=True, plot=False, warnings=True, *args, **kwargs):

    test_direct_overpopulation_vs_recombined_bands(plot=plot)
    test_3Tvib_vs_1Tvib(plot=plot)

    return True


if __name__ == '__main__':
    printm('test_overp.py:', run_testcases(plot=True))



# -*- coding: utf-8 -*-
"""
Created on Mon Nov 20 09:36:25 2017

@author: erwan

Examples
--------

Run all tests:

>>> pytest       (in command line, in project folder)

Run only fast tests (i.e: tests that have a 'fast' label)

>>> pytest -m fast
          

-------------------------------------------------------------------------------

  
"""

from __future__ import unicode_literals, print_function, absolute_import, division
from radis.lbl import SpectrumFactory, LevelsList
from radis.lbl.calc import calc_spectrum
from radis.misc.printer import printm
from radis.phys.blackbody import sPlanck
from radis.misc.utils import DatabankNotFound
from radis.test.utils import IgnoreMissingDatabase
import numpy as np
from os.path import basename
import pytest

fig_prefix = basename(__file__)+': '

# %% Tests
# ----------------


@pytest.mark.fast
def test_sPlanck_conversions(verbose=True, *args, **kwargs):

    if verbose:
        printm('Testing sPlanck conversions: ')

    s_cm = sPlanck(1000, 10000, T=1500, eps=0.3)
    I_cm2cm = s_cm.get('radiance_noslit', Iunit='mW/cm2/sr/cm_1')[1]
    I_cm2nm = s_cm.get('radiance_noslit', Iunit='mW/cm2/sr/nm')[1]

    s_nm = sPlanck(1000, 10000, T=1500, eps=0.3)
    I_nm2nm = s_nm.get('radiance_noslit', Iunit='mW/cm2/sr/nm')[1]
    I_nm2cm = s_nm.get('radiance_noslit', Iunit='mW/cm2/sr/cm_1')[1]

    assert np.allclose(I_cm2cm, I_nm2cm)
    assert np.allclose(I_nm2nm, I_cm2nm)


#@pytest.mark.needs_config_file
#@pytest.mark.needs_db_HITEMP_CO2_DUNHAM
@pytest.mark.needs_connection
def test_calc_spectrum(verbose=True, plot=True, warnings=True,
                       *args, **kwargs):
    ''' Basic example, used as a non-regression test

    Notes
    -----
    
    How long it tooks to calculate this Spectrum?

    Performance test on old NeQ package, with the [CDSD-HITEMP-JMIN] databank.
    See the caveats in the E. Pannier "Limits of CO2 NonEquilibrium Models" paper.
    (just used here as a performance monitoring)
    
    - neq 0.9.20: 18.7s

    - neq 0.9.20*: 15.4s   (removed 2nd loop on 1st isotope because of groupby().apply())

    - neq 0.9.20**: 11.7s  (after replacing fill_Evib with map() ) 

    - neq 0.9.21: 9.4s     (improve Qrot / nrot fetching performance) 

    - neq 0.9.22: 8.4s     
    
    Starting from RADIS 1.0.1, the test is run on [HITRAN-2016]_, which 
    is not valid for these temperatures but can be more conveniently 
    downloaded automatically and thus executed everytime with [Travis]_
    
    (we also expect the test to be much faster than above, but that's just 
    because the database is smaller!)
    
    - radis 0.9.20 : 
    
    '''

    if verbose:
        printm('Testing calc_spectrum match reference')

    if plot:  # Make sure matplotlib is interactive so that test are not stuck in pytest
        import matplotlib.pyplot as plt
        plt.ion()

    try:
        s = calc_spectrum(wavelength_min=4165, wavelength_max=4200,
#                          databank='CDSD-HITEMP-JMIN',
                          databank='fetch',    # not appropriate for these temperatures, but convenient for automatic testing
                          Tgas=300,
                          Tvib=1700,
                          Trot=1550,
                          path_length=0.1,
                          mole_fraction=0.5,
                          molecule='CO2',
                          isotope='1,2',
                          wstep=0.01,
                          cutoff=1e-25,
                          use_cached=True,
                          medium='vacuum',
                          verbose=verbose,
                          warnings={'MissingSelfBroadeningWarning':'ignore',
                                    'NegativeEnergiesWarning':'ignore',
                                    'HighTemperatureWarning':'ignore'})
        s.apply_slit((2, 2.5), 'nm', shape='trapezoidal')

        if plot:
            s.plot(wunit='nm')

        w, I = s.get('radiance', wunit='nm')
        w_ref = w[::100]
        # Compare against hardcoded results (neq 0.9.22, 28/06/18)
#        I_ref = np.array([0.28694463, 0.29141711, 0.32461613, 0.32909566, 0.21939511, 0.18606445,
#                          0.19740763, 0.16948599, 0.16780345, 0.15572173, 0.16770853, 0.14966064,
#                          0.13041356, 0.11751016, 0.10818072, 0.11592531, 0.04666677, 0.00177108,
#                          0.00069339])
        # Harcoded results changed for RADIS  with the change of 
        # database (HITEMP-2010 -> HITRAN-2016) and of Tvib model 
        # CDSD with (P,C,Jmin,N) in CDSD polyad -> RADIS built-in constants)
#        I_ref = np.array([ 0.29148768,  0.29646856,  0.32999337,  0.32249701,  0.2078451 ,
#                          0.18974631,  0.2019285 ,  0.17346687,  0.17211401,  0.15939359,
#                          0.17240575,  0.15395179,  0.13374185,  0.11997065,  0.10858693,
#                          0.11114162,  0.04575873,  0.00163863,  0.00062654])
        # Updated again in RADIS 0.9.20 (19/08/19) to account for the use of DLM (difference
        # not significant)
        I_ref = np.array([ 0.29060991,  0.29756722,  0.32972058,  0.3206278 ,  0.20696867,
                           0.19218358,  0.20155747,  0.17336405,  0.17218653,  0.1589136 ,
                           0.17110649,  0.15403513,  0.13376804,  0.11932659,  0.10882006,
                           0.11112725,  0.0458288 ,  0.00247956,  0.00144128])
        if plot:
            plt.plot(w_ref, I_ref, 'or', label='ref')
            plt.legend()
        assert np.allclose(I[::100], I_ref, atol=1e-6)

        return True

    except DatabankNotFound as err:
        assert IgnoreMissingDatabase(err, __file__, warnings)


#@pytest.mark.needs_db_CDSD_HITEMP_PCN
@pytest.mark.needs_connection
def test_calc_spectrum_overpopulations(verbose=True, plot=False, warnings=True,
                                       *args, **kwargs):
    ''' Non-regression test: 
        
    Example using overpopulation of the 001 asymmetric stretch first level of CO2,
    which is written (p,c,N) = (3,1,4) in [CDSD-4000]_ notation
    
    Notes
    -----
    
    In old Neq package (before RADIS):
    
    the test uses a CDSD-PCN notation for vibrational energy assignation, i.e,
    Evib = minimal energy of a (p,c,N) polyad. See the discussion on the implications
    in the E. Pannier "Limits of CO2 NonEquilibrium models" paper. 
    Better use the assignation scheme suggested in the paper. 
    But it's okay here as a non-regression test.
    
    Starting from RADIS 1.0.1, the test is run on [HITRAN-2016]_, which 
    is not valid for these temperatures but can be more conveniently 
    downloaded automatically and thus executed everytime with [Travis]_
    
    '''

    if plot:  # Make sure matplotlib is interactive so that test are not stuck in pytest
        import matplotlib.pyplot as plt
        plt.ion()

    try:
        s = calc_spectrum(wavelength_min=4165, wavelength_max=4200,
#                          databank='CDSD-HITEMP-PCN',
                          databank='fetch',    # not appropriate for these temperatures, but convenient for automatic testings
                          Tgas=300,
                          Tvib=1700,
                          Trot=1550,
#                          overpopulation={'(3,1,4)': 3},  # 00'0'1 in spectroscopic notation
                          overpopulation={'(0,0,0,1)': 3},  # 00'0'1 in spectroscopic notation
                          path_length=0.1,
                          mole_fraction=0.5,
                          molecule='CO2',
                          isotope='1,2',
                          wstep=0.01,
                          cutoff=1e-25,
                          use_cached=True,
                          medium='vacuum',
                          verbose=verbose,
                          warnings={'MissingSelfBroadeningWarning':'ignore',
                                    'NegativeEnergiesWarning':'ignore',
                                    'HighTemperatureWarning':'ignore'})
        s.apply_slit((2, 2.5), 'nm', shape='trapezoidal')

        if plot:
            s.plot()

        w, I = s.get('radiance', wunit='nm')
        w_ref = w[::100]
        # Compare against hardcoded results (neq 0.9.22, 28/06/18)
#        I_ref = np.array([0.61826008, 0.65598262, 0.79760003, 0.7958013 , 0.5792486 ,
#                          0.56727691, 0.60361258, 0.51549598, 0.51012651, 0.47133131,
#                          0.50770568, 0.45093953, 0.39129824, 0.35125324, 0.32238316,
#                          0.34542781, 0.13908073, 0.00506012, 0.00189535])
        # Harcoded results changed for RADIS v1.0.1  with the change of 
        # database (HITEMP-2010 -> HITRAN-2016) and of Tvib model 
        # CDSD with (P,C,Jmin,N) in CDSD polyad -> RADIS built-in constants)
        # 
#        I_ref = np.array([ 0.62299838,  0.66229013,  0.81037059,  0.79899315,  0.57215806,
#                          0.57626389,  0.61424273,  0.52454807,  0.5200812 ,  0.47920924,
#                          0.51843533,  0.46058817,  0.3983277 ,  0.35582979,  0.32095204,
#                          0.32821575,  0.13525543,  0.00469489,  0.00174166])
        # Updated again in RADIS 0.9.20 (16/08/19) to account for the use of DLM (difference
        # not significant)
        I_ref = np.array([ 0.62134142,  0.66722021,  0.81016539,  0.79387937,  0.56974945,
                           0.58280035,  0.6120114 ,  0.52319075,  0.5193041 ,  0.47686282,
                           0.51374777,  0.46022548,  0.3979033 ,  0.3534643 ,  0.32129239,
                           0.32786479,  0.1351593 ,  0.0068877 ,  0.00387545])
        if plot:
            plt.plot(w_ref, I_ref, 'or', label='ref')
            plt.legend()
            s.plot_populations()

        assert np.allclose(I[::100], I_ref, atol=1e-6)

        if verbose:
            printm('Test overpopulations: OK')
            
        return True

    except DatabankNotFound as err:
        assert IgnoreMissingDatabase(err, __file__, warnings)


@pytest.mark.needs_config_file
@pytest.mark.needs_db_CDSD_HITEMP_PC
#@pytest.mark.needs_connection
def test_all_calc_methods(verbose=True, plot=False, warnings=True, rtol=1e-3,
                          *args, **kwargs):
    ''' Test same spectrum for 3 different calculation variants (equilibrium, 
    non-equilibrium, per band and recombine 
    '''

    if plot:  # Make sure matplotlib is interactive so that test are not stuck in pytest
        import matplotlib.pyplot as plt
        plt.ion()

    try:

        Tgas = 1500

        iso = 1
        sf = SpectrumFactory(
            wavelength_min=4170,
            wavelength_max=4175,
            mole_fraction=1,
            path_length=0.025,
            cutoff=1e-25,
            molecule='CO2',
            isotope=iso,
            db_use_cached=True,
            lvl_use_cached=True,
            verbose=verbose)
        sf.warnings['MissingSelfBroadeningWarning'] = 'ignore'
        sf.warnings['NegativeEnergiesWarning'] = 'ignore'
        sf.warnings['HighTemperatureWarning'] = 'ignore'
#        sf.fetch_databank()   # uses HITRAN: not really valid at this temperature, but runs on all machines without install
        sf.load_databank('CDSD-HITEMP-PC')

        s_bands = sf.non_eq_bands(Tvib=Tgas, Trot=Tgas)
        lvl = LevelsList(sf.parsum_calc['CO2'][iso]
                         ['X'], s_bands, sf.params.levelsfmt)
        s_bd = lvl.non_eq_spectrum(Tvib=Tgas, Trot=Tgas)

        s_nq = sf.non_eq_spectrum(Tvib=Tgas, Trot=Tgas)
        s_eq = sf.eq_spectrum(Tgas=Tgas)

        #
        if plot:
            fig = plt.figure(fig_prefix+'Compare all calc methods')
            s_bd.plot(nfig=fig.number, color='b',
                      lw=5, label='from bands code')
            s_nq.plot(nfig=fig.number, lw=3, label='non eq code')
            s_eq.plot(nfig=fig.number, lw=2, color='r',
                      label='equilibrum code')
            plt.legend()

        assert np.isclose(s_bd.get_power(), s_nq.get_power(), rtol=rtol)
        assert np.isclose(s_bd.get_power(), s_eq.get_power(), rtol=rtol)

        if verbose:
            printm('Eq == non-eq:\t', np.isclose(s_eq.get_power(), s_nq.get_power(),
                                                 rtol=rtol))
            printm('Bands == Non-eq:\t', np.isclose(s_bd.get_power(), s_nq.get_power(),
                                                    rtol=rtol))

        if verbose:
            printm('Test all methods comparison: OK')

        return True

    except DatabankNotFound as err:
        assert IgnoreMissingDatabase(err, __file__, warnings)


@pytest.mark.needs_connection
def test_eq_vs_noneq_isotope(verbose=True, plot=False, warnings=True,
                             *args, **kwargs):
    ''' Test same spectrum for 2 different calculation codes (equilibrium, 
    non-equilibrium) in the presence of isotopes 
    
    Notes
    -----
    
    On the old NeQ package the test used [HITEMP-2010]_
    
    Starting from RADIS 1.0.1, the test is run on [HITRAN-2016]_, which 
    is not valid for these temperatures but can be more conveniently 
    downloaded automatically and thus executed everytime with [Travis]_
    
    '''

    try:
        Tgas = 1500

        sf = SpectrumFactory(
            wavelength_min=4250,
            wavelength_max=4350,
            mole_fraction=1,
            path_length=1,
            cutoff=1e-25,
            molecule='CO2',
            isotope='1,2',
            db_use_cached=True,
            verbose=verbose)
        sf.warnings['MissingSelfBroadeningWarning'] = 'ignore'
        sf.warnings['NegativeEnergiesWarning'] = 'ignore'
        sf.warnings['HighTemperatureWarning'] = 'ignore'
        sf.fetch_databank()   # uses HITRAN: not really valid at this temperature, but runs on all machines without install
#        sf.load_databank('HITEMP-CO2-DUNHAM')
        s_nq = sf.non_eq_spectrum(Tvib=Tgas, Trot=Tgas, name='Non-eq')
        s_eq = sf.eq_spectrum(Tgas=Tgas, name='Eq')

        rtol = 5e-3    # 2nd isotope calculated with placeholder energies
        match_eq_vs_non_eq = s_eq.compare_with(s_nq, spectra_only='abscoeff', rtol=rtol, plot=plot)
        match_eq_vs_non_eq *= s_eq.compare_with(s_nq, spectra_only='radiance_noslit', rtol=rtol, plot=plot)

        if verbose:
            printm('Tested eq vs non-eq (<{0:.1f}% error) with isotopes: {1}'.format(
                rtol*100, bool(match_eq_vs_non_eq)))

        assert match_eq_vs_non_eq

    except DatabankNotFound as err:
        assert IgnoreMissingDatabase(err, __file__, warnings)


def _run_testcases(plot=False, verbose=True, warnings=True, *args, **kwargs):

    # Test sPlanck and conversion functions
    test_sPlanck_conversions()

    # Test calc_spectrum function
    test_calc_spectrum()

    # Test calc_spectrum with overpopulation
    test_calc_spectrum_overpopulations(verbose=verbose, plot=plot,
                                       warnings=warnings, *args, **kwargs)

    # Compare all calc methods
    test_all_calc_methods(verbose=verbose, plot=plot,
                          warnings=warnings, *args, **kwargs)

    # Compare same spectrum with two calculation methods
    test_eq_vs_noneq_isotope(verbose=verbose, plot=plot,
                             warnings=warnings, *args, **kwargs)

    return True



# --------------------------
if __name__ == '__main__':

    printm('Testing calc.py: ', _run_testcases(verbose=True))



# -*- coding: utf-8 -*-
"""
Created on Tue Jan  9 17:14:01 2018

@author: erwan

Test broadening against HAPI and tabulated data

We're looking at CO(0->1) line 'R1' at 2150.86 cm-1

"""

from __future__ import unicode_literals, print_function, absolute_import, division
from radis.lbl.factory import SpectrumFactory
from radis.spectrum.spectrum import Spectrum
from radis import plot_diff, get_residual_integral, get_residual
from radis.misc.utils import DatabankNotFound
from radis.test.utils import IgnoreMissingDatabase, setup_test_line_databases
from radis.misc.printer import printm
from os.path import join, dirname
from numpy import isclose
import matplotlib.pyplot as plt
import pytest


@pytest.mark.fast
@pytest.mark.needs_connection
def test_broadening_vs_hapi(rtol=1e-2, verbose=True, plot=False, *args, **kwargs):
    '''
    Test broadening against HAPI and tabulated data

    We're looking at CO(0->1) line 'R1' at 2150.86 cm-1
    '''
    from radis.io.hapi import db_begin, fetch, tableList, absorptionCoefficient_Voigt

    if plot:  # Make sure matplotlib is interactive so that test are not stuck in pytest
        plt.ion()

    setup_test_line_databases()  # add HITRAN-CO-TEST in ~/.radis if not there

    # Conditions
    T = 3000
    p = 0.0001
    wstep = 0.001
    wmin = 2150  # cm-1
    wmax = 2152  # cm-1
    broadening_max_width = 10  # cm-1

    # %% HITRAN calculation
    # -----------

    # Generate HAPI database locally
    db_begin(join(dirname(__file__), __file__.replace('.py', '_HAPIdata')))
    if not 'CO' in tableList():   # only if data not downloaded already
        fetch('CO', 5, 1, wmin-broadening_max_width /
              2, wmax+broadening_max_width/2)
        # HAPI doesnt correct for side effects

    # Calculate with HAPI
    nu, coef = absorptionCoefficient_Voigt(SourceTables='CO',
                                           Environment={'T': T,  # K
                                                        'p': p/1.01325,  # atm
                                                        },
                                           WavenumberStep=wstep,
                                           HITRAN_units=False)

    s_hapi = Spectrum.from_array(nu, coef, 'abscoeff', 'cm-1', 'cm_1',
                                 conditions={'Tgas': T},
                                 name='HAPI')

    # %% Calculate with RADIS
    # ----------
    sf = SpectrumFactory(
        wavenum_min=wmin,
        wavenum_max=wmax,
        mole_fraction=1,
        path_length=1,   # doesnt change anything
        wstep=wstep,
        pressure=p,
        broadening_max_width=broadening_max_width,
        isotope=[1],
        warnings={'MissingSelfBroadeningWarning':'ignore',
                  'NegativeEnergiesWarning':'ignore',
                  'HighTemperatureWarning':'ignore',
                  'GaussianBroadeningWarning':'ignore'}
        )  # 0.2)
    sf.load_databank('HITRAN-CO-TEST')
#    s = pl.non_eq_spectrum(Tvib=T, Trot=T, Ttrans=T)
    s = sf.eq_spectrum(Tgas=T, name='RADIS')
    
    
    if plot:  # plot broadening of line of largest linestrength
        sf.plot_broadening(i=sf.df1.S.idxmax())

    # Plot and compare
    res = abs(get_residual_integral(s, s_hapi, 'abscoeff'))
    if plot:
        plot_diff(s, s_hapi, var='abscoeff',
                  title='{0} bar, {1} K (residual {2:.2g}%)'.format(p, T, res*100), show_points=False)
        plt.xlim((wmin, wmax))
    if verbose:
        printm('residual:', res)
    assert res < rtol


@pytest.mark.fast
def test_broadening_methods_different_conditions(verbose=True, plot=False, *args, **kwargs):
    '''
    Test direct Voigt broadening vs convolution of Gaussian x Lorentzian
    for different spectral grid resolution
    
    Notes
    ----- 
    
    Reference broadening calculated manually with the HWHM formula of 
    `HITRAN.org <https://hitran.org/docs/definitions-and-units/>`_
    '''

    if plot:  # Make sure matplotlib is interactive so that test are not stuck in pytest
        plt.ion()

    setup_test_line_databases()  # add HITRAN-CO-TEST in ~/.radis if not there

    # Conditions
    wstep = 0.005
    wmin = 2150.4  # cm-1
    wmax = 2151.4  # cm-1
    broadening_max_width = 2  # cm-1
    
    
    for (T, p, fwhm_lorentz, fwhm_gauss) in [
         # K, bar, expected FWHM for Lotentz, gauss (cm-1)
         (3000, 1,      0.02849411,      0.01594728),
         (300,  1,      0.16023415,      0.00504297),
         (3000, 0.01,   0.00028494,      0.01594728),
         ]:

        # %% Calculate with RADIS
        # ----------
        sf = SpectrumFactory(
            wavenum_min=wmin,
            wavenum_max=wmax,
            mole_fraction=1,
            path_length=1,   # doesnt change anything
            wstep=wstep,
            pressure=p,
            broadening_max_width=broadening_max_width,
            isotope='1',
            verbose=False,
            warnings={'MissingSelfBroadeningWarning':'ignore',
                      'NegativeEnergiesWarning':'ignore',
                      'HighTemperatureWarning':'ignore',
                      'OutOfRangeLinesWarning':'ignore',
                      'GaussianBroadeningWarning':'ignore',
                      'CollisionalBroadeningWarning':'ignore'}
            )
        sf.load_databank('HITRAN-CO-TEST')
        # Manually filter line database, keep one line only:
        sf.df0.drop(sf.df0[sf.df0.vu!=1].index, inplace=True)
        assert isclose(sf.df0.wav, 2150.856008)
        
        # Calculate spectra (different broadening methods)
        sf._broadening_method = 'voigt'
        s_voigt = sf.eq_spectrum(Tgas=T, name='direct')
        
        # assert broadening FWHM are correct
        assert isclose(2*float(sf.df1.hwhm_gauss), fwhm_gauss)
        assert isclose(2*float(sf.df1.hwhm_lorentz), fwhm_lorentz)
        
        
        sf._broadening_method = 'convolve'
        s_convolve = sf.eq_spectrum(Tgas=T, name='convolve')
    
        # assert broadening FWHM are correct
        assert isclose(2*float(sf.df1.hwhm_gauss), fwhm_gauss)
        assert isclose(2*float(sf.df1.hwhm_lorentz), fwhm_lorentz)
        
        
        res = get_residual(s_voigt, s_convolve, 'abscoeff')
        
        if verbose:
            print('{0} K, {1} bar: FWHM lorentz = {2:.3f} cm-1, FWHM gauss = {3:.3f} cm-1'.format(
                    T, p, 2*float(sf.df1.hwhm_lorentz), 2*float(sf.df1.hwhm_gauss)))

        if plot:
            plot_diff(s_voigt, s_convolve, 'abscoeff', 
                      title=r'T {0} K, p {1} bar: w$_\mathrm{{L}}$ {2:.3f}, w$_\mathrm{{G}}$ {3:.3f} cm$^{{-1}}$'.format(
                      T, p, 2*float(sf.df1.hwhm_lorentz), float(sf.df1.hwhm_gauss)))
        
        # assert all broadening methods match
        assert res < 2e-4
        
        
@pytest.mark.fast
def test_broadening_methods_different_wstep(verbose=True, plot=False, *args, **kwargs):
    '''
    Test direct Voigt broadening vs convolution of Gaussian x Lorentzian
    for different spectral grid resolution
    '''

    if plot:  # Make sure matplotlib is interactive so that test are not stuck in pytest
        plt.ion()

    setup_test_line_databases()  # add HITRAN-CO-TEST in ~/.radis if not there

    # Conditions
    T = 3000
    p = 1
    wmin = 2150  # cm-1
    wmax = 2152  # cm-1
    broadening_max_width = 10  # cm-1
    
    for i, wstep in enumerate([0.01, 0.1, 0.5]):

        # %% Calculate with RADIS
        # ----------
        sf = SpectrumFactory(
            wavenum_min=wmin,
            wavenum_max=wmax,
            mole_fraction=1,
            path_length=1,   # doesnt change anything
            wstep=wstep,
            pressure=p,
            broadening_max_width=broadening_max_width,
            isotope='1',
            verbose=False,
            warnings={'MissingSelfBroadeningWarning':'ignore',
                      'NegativeEnergiesWarning':'ignore',
                      'HighTemperatureWarning':'ignore',
                      'GaussianBroadeningWarning':'ignore'}
            )  # 0.2)
        sf.load_databank('HITRAN-CO-TEST')
    #    s = pl.non_eq_spectrum(Tvib=T, Trot=T, Ttrans=T)
        sf._broadening_method = 'voigt'
        s_voigt = sf.eq_spectrum(Tgas=T, name='direct')
        
        sf._broadening_method = 'convolve'
        s_convolve = sf.eq_spectrum(Tgas=T, name='convolve')
    
        res = get_residual(s_voigt, s_convolve, 'abscoeff')
        
        if verbose:
            print('Residual:', res)

        # plot the last one    
        if plot:
            plot_diff(s_voigt, s_convolve, 'abscoeff', nfig='test_voigt_broadening_methods'+str(i),
                      title='P {0} bar, T {1} K, wstep {2} cm-1'.format(p, T, wstep))
        
        assert res < 2e-4
        
        
@pytest.mark.fast
def test_broadening_DLM(verbose=True, plot=False, *args, **kwargs):
    '''
    Test use of lineshape template for broadening calculation. 
    
    Ensures that results are the same with and without DLM.
    '''

    if plot:  # Make sure matplotlib is interactive so that test are not stuck in pytest
        plt.ion()

    setup_test_line_databases()  # add HITRAN-CO-TEST in ~/.radis if not there

    # Conditions
    T = 3000
    p = 1
    wstep = 0.002
    wmin = 2150  # cm-1
    wmax = 2152  # cm-1
    broadening_max_width = 10  # cm-1

    # %% Calculate with RADIS
    # ----------
    sf = SpectrumFactory(
        wavenum_min=wmin,
        wavenum_max=wmax,
        mole_fraction=1,
        path_length=1,   # doesnt change anything
        wstep=wstep,
        pressure=p,
        broadening_max_width=broadening_max_width,
        isotope='1',
        verbose=False,
        warnings={'MissingSelfBroadeningWarning':'ignore',
                  'NegativeEnergiesWarning':'ignore',
                  'HighTemperatureWarning':'ignore',
                  'GaussianBroadeningWarning':'ignore'}
        )  # 0.2)
    sf.load_databank('HITRAN-CO-TEST')
    
    # Reference: calculate without DLM 
    assert sf.misc['chunksize'] is None
    s_ref = sf.eq_spectrum(Tgas=T)
    s_ref.name = 'Reference ({0:.2f}s)'.format(s_ref.conditions['calculation_time'])
    
    # DLM:
    sf.misc['chunksize'] = 'DLM'
    sf._broadening_method = 'convolve'
    s_dlm = sf.eq_spectrum(Tgas=T)
    s_dlm.name = 'DLM ({0:.2f}s)'.format(s_dlm.conditions['calculation_time'])
    # DLM Voigt with Whiting approximation:
    sf._broadening_method = 'voigt'
    s_dlm_voigt = sf.eq_spectrum(Tgas=T)
    s_dlm_voigt.name = 'DLM Whiting ({0:.2f}s)'.format(s_dlm_voigt.conditions['calculation_time'])
    
    # Compare
    res = get_residual(s_ref, s_dlm, 'abscoeff')
    res_voigt = get_residual(s_dlm, s_dlm_voigt, 'abscoeff')
    
    if verbose:
        print('Residual:', res)

    # plot the last one    
    if plot:
        plot_diff(s_ref, s_dlm, 'abscoeff')
        plt.legend()
        plot_diff(s_dlm, s_dlm_voigt, 'abscoeff')
    
    assert res < 1.2e-5
    assert res_voigt < 1e-5

@pytest.mark.fast
def test_broadening_DLM_FT(verbose=True, plot=False, *args, **kwargs):
    '''
    Test use of DLM with and without Fourier Transform
    
    Ensures that results are the same, and compare calculation times.
    '''

    if plot:  # Make sure matplotlib is interactive so that test are not stuck in pytest
        plt.ion()

    setup_test_line_databases()  # add HITRAN-CO-TEST in ~/.radis if not there

    # Conditions
    T = 3000
    p = 1
    wstep = 0.002
    wmin = 2000  # cm-1
    wmax = 2300  # cm-1
    broadening_max_width = 10  # cm-1

    # %% Calculate with RADIS
    # ----------
    sf = SpectrumFactory(
        wavenum_min=wmin,
        wavenum_max=wmax,
        mole_fraction=1,
        path_length=1,   # doesnt change anything
        wstep=wstep,
        pressure=p,
        broadening_max_width=broadening_max_width,
        isotope='1',
        verbose=verbose,
        chunksize='DLM',
        warnings={'MissingSelfBroadeningWarning':'ignore',
                  'NegativeEnergiesWarning':'ignore',
                  'HighTemperatureWarning':'ignore',
                  'GaussianBroadeningWarning':'ignore'}
        )  # 0.2)
    sf.load_databank('HITRAN-CO-TEST')
    
    # DLM, real space
    if verbose: print('\nConvolve version \n')
    s_dlm = sf.eq_spectrum(Tgas=T)
    s_dlm.name = 'DLM ({0:.2f}s)'.format(s_dlm.conditions['calculation_time'])
    
    # DLM , with Fourier
    if verbose: print('\nFFT version \n')
    sf._broadening_method = 'fft'
    s_dlm_fft = sf.eq_spectrum(Tgas=T)
    s_dlm_fft.name = 'DLM FFT ({0:.2f}s)'.format(s_dlm_fft.conditions['calculation_time'])
    
    # Compare
    res = get_residual(s_dlm, s_dlm_fft, 'abscoeff')
    
    if verbose:
        print('Residual:', res)

    # plot    
    if plot:
        plot_diff(s_dlm, s_dlm_fft, 'abscoeff')
        plt.legend()
        
    assert res < 5e-6

@pytest.mark.fast
def test_broadening_DLM_noneq(verbose=True, plot=False, *args, **kwargs):
    '''
    Test Noneq version of DLM and makes sure it gives the same results as the eq
    one when used with Tvib=Trot 
    
    '''

    if plot:  # Make sure matplotlib is interactive so that test are not stuck in pytest
        plt.ion()

    setup_test_line_databases()  # add HITRAN-CO2-TEST in ~/.radis if not there

    # Conditions
    p = 1
    wstep = 0.002
    wmin = 2380  # cm-1
    wmax = 2400  # cm-1
    broadening_max_width = 10  # cm-1

    # %% Calculate with RADIS
    # ----------
    sf = SpectrumFactory(
        wavenum_min=wmin,
        wavenum_max=wmax,
        mole_fraction=1,
        path_length=1,   # doesnt change anything
        wstep=wstep,
        pressure=p,
        broadening_max_width=broadening_max_width,
        isotope='1',
        verbose=3,
        warnings={'MissingSelfBroadeningWarning':'ignore',
                  'NegativeEnergiesWarning':'ignore',
                  'HighTemperatureWarning':'ignore',
                  'GaussianBroadeningWarning':'ignore'}
        )  # 0.2)
    sf.load_databank('HITRAN-CO2-TEST')
    
    # DLM:
    sf.misc['chunksize'] = 'DLM'
    s_dlm_eq = sf.eq_spectrum(Tgas=3000)
    s_dlm_eq.name = 'DLM eq ({0:.2f}s)'.format(s_dlm_eq.conditions['calculation_time'])
    
    s_dlm_noneq = sf.non_eq_spectrum(Tvib=3000, Trot=3000)
    s_dlm_noneq.name = 'DLM noneq ({0:.2f}s)'.format(s_dlm_noneq.conditions['calculation_time'])
    
    # Compare
    res = get_residual(s_dlm_eq, s_dlm_noneq, 'radiance_noslit')
    
    if verbose:
        print('Residual:', res)

    # plot
    if plot:
        plot_diff(s_dlm_eq, s_dlm_noneq)
    
    assert res <= 4e-5 



#@pytest.mark.needs_config_file
#@pytest.mark.needs_db_HITEMP_CO2_DUNHAM
@pytest.mark.needs_connection
def test_abscoeff_continuum(plot=False, verbose=2, warnings=True, threshold=0.1, *args, **kwargs):
    ''' 
    Test calculation with pseudo-continuum

    Assert results on abscoeff dont change
    
    
    Notes
    -----
    
    Uses HITRAN so it can deployed and tested on [Travis]_, but we should switch
    to HITEMP if some HITEMP files can be downloaded automatically at the 
    execution time. 

    '''

    if plot:  # Make sure matplotlib is interactive so that test are not stuck in pytest
        plt.ion()

    try:
        if verbose:
            printm('>>> test_abscoeff_continuum')

        sf = SpectrumFactory(wavelength_min=4200, 
                             wavelength_max=4500,
                             parallel=False, bplot=False, cutoff=1e-23,
                             molecule='CO2',
                             isotope='1,2', 
                             db_use_cached=True,
                             broadening_max_width=10,
                             path_length=0.1,
                             mole_fraction=1e-3,
                             medium='vacuum',
                             verbose=verbose)
        sf.warnings.update({
            'MissingSelfBroadeningWarning': 'ignore',
            'NegativeEnergiesWarning': 'ignore',
            'LinestrengthCutoffWarning': 'ignore',
            'HighTemperatureWarning':'ignore'})
        sf.fetch_databank()   # uses HITRAN: not really valid at this temperature, but runs on all machines without install
#        sf.load_databank('HITEMP-CO2-DUNHAM')       # to take a real advantage of abscoeff continuum, should calculate with HITEMP
        sf._export_continuum = True   # activate it 

        # Calculate one without pseudo-continuum
        sf.params.pseudo_continuum_threshold = 0
        s1 = sf.eq_spectrum(Tgas=2000)
        s1.name = 'All lines resolved ({0}) ({1:.1f}s)'.format(s1.conditions['lines_calculated'],
                                                               s1.conditions['calculation_time'])
        assert s1.conditions['pseudo_continuum_threshold'] == 0

        # Calculate one with pseudo-continuum
        sf.params.pseudo_continuum_threshold = threshold
        s2 = sf.eq_spectrum(Tgas=2000)
        s2.name = 'Semi-continuum + {0} lines ({1:.1f}s)'.format(s2.conditions['lines_calculated'],
                                                                 s2.conditions['calculation_time'])
        assert s2.conditions['pseudo_continuum_threshold'] == threshold
        assert 'abscoeff_continuum' in s2.get_vars()

        # Plot
        if plot:
            plot_diff(s1, s2, 'radiance_noslit', Iunit='µW/cm2/sr/nm',
                      nfig='test_abscoeff_continuum: diff')

            s2.plot('abscoeff', label='Full spectrum',
                    nfig='test_abscoeff_continuum: show continuum', force=True)
            s2.plot('abscoeff_continuum', nfig='same', label='Pseudo-continuum'.format(
                    s2.conditions['lines_in_continuum']),
                    force=True)
            plt.legend()

        # Compare
        res = get_residual(s1, s2, 'abscoeff')
        if verbose:
            printm('residual:', res)

        globals().update(locals())

        assert res < 1.32e-6

    except DatabankNotFound as err:
        assert IgnoreMissingDatabase(err, __file__, warnings)

#@pytest.mark.needs_config_file
#@pytest.mark.needs_db_HITEMP_CO2_DUNHAM
@pytest.mark.needs_connection
def test_noneq_continuum(plot=False, verbose=2, warnings=True, *args, **kwargs):
    ''' 
    Test calculation with pseudo-continuum under nonequilibrium

    Assert results on emisscoeff dont change

    
    Notes
    -----
    
    Uses HITRAN so it can deployed and tested on [Travis]_, but we should switch
    to HITEMP if some HITEMP files can be downloaded automatically at the 
    execution time. 

    '''
   
    if plot:  # Make sure matplotlib is interactive so that test are not stuck in pytest
        plt.ion()

    try:
        if verbose:
            printm('>>> test_noneq_continuum')

        sf = SpectrumFactory(wavelength_min=4200, wavelength_max=4500,
                             parallel=False, bplot=False, cutoff=1e-23,
                             molecule='CO2',
                             isotope='1,2', db_use_cached=True,
                             broadening_max_width=10,
                             path_length=0.1,
                             mole_fraction=1e-3,
                             medium='vacuum',
                             verbose=verbose)
        sf.warnings.update({
            'MissingSelfBroadeningWarning': 'ignore',
            'NegativeEnergiesWarning': 'ignore',
            'LinestrengthCutoffWarning': 'ignore',
            'HighTemperatureWarning':'ignore'})
        sf.fetch_databank()   # uses HITRAN: not really valid at this temperature, but runs on all machines without install
#        sf.load_databank('HITEMP-CO2-DUNHAM')       # to take a real advantage of abscoeff continuum, should calculate with HITEMP
        sf._export_continuum = True   # activate it 

        # Calculate one without pseudo-continuum
        sf.params.pseudo_continuum_threshold = 0
        s1 = sf.non_eq_spectrum(Tvib=2000, Trot=1000)
        s1.name = 'All lines resolved ({0}) ({1:.1f}s)'.format(s1.conditions['lines_calculated'],
                                                               s1.conditions['calculation_time'])
        assert s1.conditions['pseudo_continuum_threshold'] == 0

        # Calculate one with pseudo-continuum
        sf.params.pseudo_continuum_threshold = 0.05
        s2 = sf.non_eq_spectrum(Tvib=2000, Trot=1000)
        s2.name = 'Semi-continuum + {0} lines ({1:.1f}s)'.format(s2.conditions['lines_calculated'],
                                                                 s2.conditions['calculation_time'])
        assert s2.conditions['pseudo_continuum_threshold'] == 0.05
        assert 'abscoeff_continuum' in s2.get_vars()
        assert 'emisscoeff_continuum' in s2.get_vars()

        # Plot
        if plot:
            plot_diff(s1, s2, 'radiance_noslit', Iunit='µW/cm2/sr/nm',
                      nfig='test_noneq_continuum: diff')

            plt.figure('test_noneq_continuum: show continuum').clear()
            s2.plot('emisscoeff', label=s2.name,
                    nfig='test_noneq_continuum: show continuum')
            s2.plot('emisscoeff_continuum', nfig='same', label='Pseudo-continuum (aggreg. {0:g} lines)'.format(
                    s2.conditions['lines_in_continuum']), force=True)

        # Compare
        res = get_residual(s1, s2, 'abscoeff') + get_residual(s1, s2, 'emisscoeff')
        if verbose:
            printm('residual:', res)

        assert res < 5.2e-6

    except DatabankNotFound as err:
        assert IgnoreMissingDatabase(err, __file__, warnings)



def _run_testcases(plot=False, verbose=True, *args, **kwargs):

   # Test broadening
    test_broadening_vs_hapi(plot=plot, verbose=verbose, *args, **kwargs)
    test_broadening_methods_different_conditions(plot=plot, verbose=verbose, *args, **kwargs)
    test_broadening_methods_different_wstep(plot=plot, verbose=verbose, *args, **kwargs)
    test_broadening_DLM(plot=plot, verbose=verbose, *args, **kwargs)
    test_broadening_DLM_FT(plot=plot, verbose=3, *args, **kwargs)
    test_broadening_DLM_noneq(plot=plot, verbose=verbose, *args, **kwargs)

    # Test pseudo-continuum
    test_abscoeff_continuum(plot=plot, verbose=verbose, *args, **kwargs)
    test_noneq_continuum(plot=plot, verbose=verbose, *args, **kwargs)
    
    return True


if __name__ == '__main__':
    printm('test_broadening: ', _run_testcases(plot=True, verbose=True, debug=False))



# -*- coding: utf-8 -*-
"""
Created on Mon May 22 18:23:18 2017

@author: erwan

Build a Parallel Spectrum Factory

Summary
-------

These loads the database once, then calculates a given number of cases in 
parallel using a dedicated SpectrumFactory. Each case is calculated on one 
core only. This is more suited to calculations of large number of small spectra. 
If you want to calculate only one large spectra consider using a single SpectrumFactory
in parallel mode

Routine Listing
---------------

Most methods are written in inherited class with the following inheritance scheme:
    
:py:class:`~radis.lbl.loader.DatabankLoader` > :py:class:`~radis.lbl.base.BaseFactory` > 
:py:class:`~radis.lbl.broadening.BroadenFactory` > :py:class:`~radis.lbl.bands.BandFactory` > 
:py:class:`~radis.lbl.factory.SpectrumFactory` > :py:class:`~radis.lbl.parallel.ParallelFactory`

Examples
--------

Just replace SpectrumFactory by ParallelFactory in your codes. And have a list
as an input of `eq_spectrum` or `non_eq_spectrum`. 

See Also
--------

Refer to :class:`~radis.lbl.factory.SpectrumFactory` for more information. 

-------------------------------------------------------------------------------

"""

from __future__ import absolute_import, print_function, unicode_literals, division
from radis.lbl import SpectrumFactory
from radis.misc.basics import is_list, is_float
from radis.misc.utils import FileNotFoundError
from multiprocessing import Pool, cpu_count
#from multiprocessing.pool import ThreadPool
import sys
from time import time
import numpy as np
from copy import deepcopy
from warnings import warn
import os
from six.moves import zip
from uuid import uuid1
from radis.lbl.loader import df_metadata
from radis.misc.basics import expand_metadata


def _distribute_eq_spectrum(args):
    ''' Clone the Factory and calculate eq_spectrum over several clones '''
    cast_factory, Tgas, mole_fraction, path_length = args
    # ... (dev) must match p.map(_distribute_eq_spectrum...) order
    factory = deepcopy(cast_factory)
    # Update id
    factory._id = uuid1()
    return SpectrumFactory.eq_spectrum(factory, Tgas, mole_fraction=mole_fraction, 
                                       path_length=path_length)


def _distribute_noneq_spectrum(args):
    ''' Clone the Factory and calculate non_eq_spectrum over several clones '''
    cast_factory, Tvib, Trot, mole_fraction, path_length = args
    # ... (dev) must match p.map(_distribute_noneq_spectrum...) order
    factory = deepcopy(cast_factory)
    # Update id
    factory._id = uuid1()
    return SpectrumFactory.non_eq_spectrum(factory, Tvib, Trot, mole_fraction=mole_fraction, 
                                           path_length=path_length)


class ParallelFactory(SpectrumFactory):
    ''' A Parallel version of :class:`~radis.lbl.parallel.SpectrumFactory`. 
    Use `Nprocs` as argument to change the number of processors to be used.


    Notes
    -----

    for Windows users:

    On Windows the subprocesses will import (i.e. execute) the main module at start. 
    You need to protect the main code like this to avoid creating subprocesses recursively::

        from radis.lbl import ParallelFactory
        if __name__ == '__main__':    
            # your code

    More info: https://stackoverflow.com/questions/18204782/runtimeerror-on-windows-trying-python-multiprocessing

    Parallel is only implemented on Python 3. 

    Parameters
    ----------

    wavenum_min: (cm-1)
        minimum wavenumber to be processed in cm^-1

    wavenum_max: (cm-1)
        maximum wavenumber to be processed in cm^-1

    Tref: (K)
        reference temperature for calculations, HITRAN database uses 296 Kelvin
        default: 3400K

    pressure: (bar)
        partial pressure of gas in bar. Default 1.01325 (1 atm)

    mole_fraction: N/D
        species mole fraction. Default 1. Note that the rest of the gas
        is considered to be air for collisional broadening.

    path_length: (cm)
        path length in cm. Default 1.

    isotope: int, str, or list of int
        isotope id (sorted by relative density). Default [1,2] (eg: CO2-626, CO2-636 for CO2)

    Other Parameters
    ----------------

    Nprocs: int
        Number of processors to use. Default total number of threads. 

    Tref: K
        Reference temperature for calculations (linestrength temperature
        correction). HITRAN database uses 296 Kelvin. Default 296 K

    broadening_max_width: cm-1
        Full width over which to compute the broadening. Large values will create
        a huge performance drop (because convolutions are not vectorized).
        Also, the calculated spectral range is increased (by broadening_max_width/2
        on each side) to take into account overlaps from out-of-range lines. 
        Default 10 cm-1.

    wstep: cm-1
        Spacing of calculated absorbance spectrum. Default 0.01 cm-1

    cutoff: float (~ unit of Linestrength: cm-1/(#.cm-2))
        discard linestrengths that are lower that this, to reduce calculation
        times. 1e-27 is what is generally used to generate databases such as 
        CDSD. If 0, no cutoff. Default 1e-27.

    bplot: boolean
        plot intermediary results (like slit function generation). Default ``False``.

    save_memory: boolean
        if ``True``, removes databases calculated by intermediate functions (for
        instance, delete the full database once the linestrength cutoff criteria
        was applied). This saves some memory but requires to reload the database
        & recalculate the linestrength for each new parameter. Default ``False``.

    chunksize: float
        Splits the lines database in several chuncks during calculation, else
        the multiplication of lines over all spectral range takes too much memory
        and slows the system down. Chuncksize let you change the default chunck
        size. Default 3e7
        
    Examples
    --------
    
    Refer to the online :ref:`Examples <label_examples>`. 


    See Also
    --------

    :class:`~radis.lbl.factory.SpectrumFactory`
    '''

    def __init__(self, *args, **kwargs):

        Nprocs = kwargs.pop('Nprocs', cpu_count())  # default to cpu_count()
        kwargs['parallel'] = False  # calculate each spectrum on one core only

        super(ParallelFactory, self).__init__(*args, **kwargs)

        self.misc.Nprocs = Nprocs

        if os.name == 'nt' and self.verbose:
            print(("Warning. On Windows, if executing the parallel versions of " +
                   "eq_spectrum and non_eq_spectrum in a script, make sure you " +
                   "embed them in `if __name__ == '__main__'` or daemons processes" +
                   " won't start"))

    def eq_spectrum(self, Tgas, mole_fraction=None, path_length=None):
        ''' Generate a spectrum at equilibrium

        Parameters
        ----------

        Tgas: list or float
            Gas temperature (K)

        mole_fraction: list or float
            database species mole fraction. If None, Factory mole fraction is used.

        path_length: list or float
            slab size (cm). If None, Factory mole fraction is used.

        Returns
        -------

        Returns a :class:`~radis.spectrum.spectrum.Spectrum` object

        Use the :meth:`~radis.spectrum.spectrum.Spectrum.get` method to get something
        among ``['radiance', 'radiance_noslit', 'absorbance', etc...]``

        Or directly the :meth:`~radis.spectrum.spectrum.Spectrum.plot` method
        to plot it

        See [1]_ to get an overview of all Spectrum methods

        References
        ----------

        .. [1] RADIS doc: `Spectrum how to? <https://radis.readthedocs.io/en/latest/spectrum/spectrum.html#label-spectrum>`__


        Notes
        -----

        Process:

        Calculate line strenghts correcting the CDSD reference one. Then call
        the main routine that sums over all lines


        '''

        args = self._format_input(**{'Tgas': Tgas, 'mole_fraction': mole_fraction,
                                     'path_length': path_length})
        Tgas = args.pop('Tgas')
        mole_fraction = args.pop('mole_fraction')
        path_length = args.pop('path_length')
        
        # Expand metadata before copy, else it will be lost. 
        # @dev: See :py:data:`~radis.lbl.loader.df_metadata` for more explanations
        expand_metadata(self.df0, [k for k in df_metadata if hasattr(self.df0, k)])

        N = len(Tgas)
        cast_factory = [self]*N  # note that this is the copy of a same object!

        Nprocs = min(self.misc.Nprocs, N)
        if self.verbose:
            print(
                ('Calculate {0} spectra on {1} procs'.format(len(Tgas), Nprocs)))
        t1 = time()
        
        try:
            with Pool(Nprocs) as p: # Python 3 only
                spec_list = p.map(_distribute_eq_spectrum, list(
                    zip(cast_factory, Tgas, mole_fraction, path_length)))
        except AttributeError:
            if sys.version_info[0] == 2:
                raise NotImplementedError('parallel is only implemented in Python 3. Time to switch!')
            else:
                raise
        
        if self.verbose:
            print(('{0} spectra calculated in {1:.1f}s'.format(N, time()-t1)))

        return spec_list

    def non_eq_spectrum(self, Tvib, Trot, mole_fraction=None, path_length=None):
        ''' Calculate emission spectrum in non-equilibrium case. Calculates
        absorption with broadened linestrength and emission with broadened
        Einstein coefficient.

        Parameters
        ----------

        Tvib: list or float
            vibrational temperature [K]

        Trot: list or float
            rotational temperature [K]

        mole_fraction: list or float
            database species mole fraction. If None, Factory mole fraction is used.

        path_length: list or float. If None, Factory mole fraction is used.
            slab size (cm)

        Returns
        -------

        Returns a :class:`~radis.spectrum.spectrum.Spectrum` object

        Use the :meth:`~radis.spectrum.spectrum.Spectrum.get` method to get something
        among ``['radiance', 'radiance_noslit', 'absorbance', etc...]``

        Or directly the :meth:`~radis.spectrum.spectrum.Spectrum.plot` method
        to plot it

        See [1]_ to get an overview of all Spectrum methods

        References
        ----------

        .. [1] RADIS doc: `Spectrum how to? <https://radis.readthedocs.io/en/latest/spectrum/spectrum.html#label-spectrum>`__


        Notes
        -----

        Hypothesis:

        - Trot = Ttrans

        '''
        args = self._format_input(**{'Tvib': Tvib, 'Trot': Trot, 'mole_fraction': mole_fraction,
                                     'path_length': path_length})
        Tvib = args['Tvib']
        Trot = args['Trot']
        mole_fraction = args['mole_fraction']
        path_length = args['path_length']
        
        # Expand metadata before copy, else it will be lost. 
        # @dev: See :py:data:`~radis.lbl.loader.df_metadata` for more explanations
        expand_metadata(self.df0, [k for k in df_metadata if hasattr(self.df0, k)])

        N = len(Tvib)
        cast_factory = [self]*N  # note that this is the copy of a same object!

        Nprocs = min(self.misc.Nprocs, N)
        if self.verbose:
            print(
                ('Calculate {0} spectra on {1} procs'.format(len(Tvib), Nprocs)))
        t1 = time()
        
        try:
            with Pool(Nprocs) as p:          # Note: Python 3 syntax only
                spec_list = p.map(_distribute_noneq_spectrum, list(zip(cast_factory, Tvib, Trot, mole_fraction, path_length)))
        except AttributeError:
            if sys.version_info[0] == 2:
                raise NotImplementedError('parallel is only implemented in Python 3. Time to switch!')
            else:
                raise
        
        if self.verbose:
            print(('{0} spectra calculated in {1:.1f}s'.format(N, time()-t1)))

        return spec_list

    def _format_input(self, **kwargs):
        ''' Format the input of parallel calculation request. Returns lists of 
        same lengths that can be parsed with zip(). User input can be lists, 
        or floats instead of constant-values list '''
        N = None   # list length if there are list involved
        kwout = {}
        for k, v in kwargs.items():
            if is_list(v):
                if N is None:
                    N = len(v)
                else:
                    if len(v) != N:
                        raise ValueError('Invalid input for {0}: '.format(k) +
                                         'all list should have the same length'
                                         '(you may use floats too)')
                kwout[k] = v

        # Now let's turn floats into list of length N
        if N is None:
            warn('Using ParallelFactory for single case')
            N = 1
        for k, v in kwargs.items():
            if type(v) in [int, np.int64]:
                v = float(v)   # int is not serializable (for some reason)
            if is_list(v):
                continue  # already done
            elif is_float(v) or v is None:
                kwout[k] = [v]*N
            else:
                raise ValueError('Invalid input for {0}: '.format(k) +
                                 'input should be list-like or float-like' +
                                 '({0} instead)'.format(type(v)))

        # check output is correct:
        for v in kwout.values():
            assert(len(v) == N)

        return kwout


# %% Main
if __name__ == '__main__':

    from radis.test.lbl.test_parallel import _run_testcases
    print(('Tested parallel.py:', _run_testcases(plot=True)))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""

Summary
-------

Test Spectrum rescaling methods

-------------------------------------------------------------------------------


"""

from __future__ import print_function, absolute_import, division, unicode_literals
import radis
from radis.spectrum.rescale import get_redundant, get_recompute
from radis.tools.database import load_spec
from radis.test.utils import getTestFile
from radis.misc.printer import printm
import numpy as np
import pytest

@pytest.mark.fast
def test_compression(verbose=True, warnings=True, *args, **kwargs):
    ''' Test that redundant quantities are properly infered from already known 
    spectral quantities '''

    # Get spectrum
    s1 = load_spec(getTestFile('CO_Tgas1500K_mole_fraction0.01.spec'), binary=True)
    
    s1.conditions['thermal_equilibrium'] = True
    s1.update()

    # Analyse redundant spectral quantities
    redundant = get_redundant(s1)
    if verbose:
        print(redundant)

    assert redundant == {'emissivity_noslit': True, 'radiance_noslit': True,
                         'emisscoeff': True, 'transmittance_noslit': True,
                         'absorbance': True, 'abscoeff': False}

    return True


@pytest.mark.fast
def test_update_transmittance(verbose=True, warnings=True, *args, **kwargs):
    ''' Test that update can correctly recompute missing quantities '''
    # TODO: add one with radiance too

    # Work with a Spectrum object that was generated by Specair
    s = load_spec(getTestFile('N2C_specair_380nm.spec'))
    w1, T1 = s.get('transmittance_noslit')   # our reference

    if verbose:
        debug_mode = radis.DEBUG_MODE  # shows all the rescale steps taken
        radis.DEBUG_MODE = True

    # Now let's apply some update() steps

    # 1) Make sure updating doesnt change anything
    s.update()
    w2, T2 = s.get('transmittance_noslit')

    # 2) Now recompute from abscoeff
    del s._q['transmittance_noslit']
    s.update()
    w2, T3 = s.get('transmittance_noslit')

    # 3) Now recompute from absorbance
    del s._q['transmittance_noslit']
    del s._q['abscoeff']
    s.update()
    w2, T4 = s.get('transmittance_noslit')

    if verbose:
        radis.DEBUG_MODE = debug_mode

    # Compare
    assert np.allclose(T1, T2)
    assert np.allclose(T1, T3)
    assert np.allclose(T1, T4)

    return True


def test_get_recompute(verbose=True, *args, **kwargs):
    ''' Make sure :func:`~radis.spectrum.rescale.get_recompute` works as expected

    Here, we check which quantities are needed to recompute radiance_noslit'''

    # Equilibrium
    # -----------
    s = load_spec(getTestFile('CO_Tgas1500K_mole_fraction0.01.spec'), binary=True)
    
    assert s.get_vars() == ['abscoeff']
    assert s.conditions['thermal_equilibrium']
    # At equilibrium, everything should be deduced from abscoeff
    assert set(get_recompute(s, ['radiance_noslit'])) == set(
        ('radiance_noslit', 'abscoeff'))

    # Non Equilibrium
    # ----------------
    s.conditions['Tvib'] = 2000  
    # a problem should be detected by is_at_equilibrium()
    with pytest.raises(AssertionError):
        assert not s.is_at_equilibrium(check='error')
    # force non equilibrium
    s.conditions['thermal_equilibrium'] = False
    
    # Now more data is needed:
    assert set(get_recompute(s, ['radiance_noslit'])) == set(
        ('abscoeff', 'emisscoeff', 'radiance_noslit'))


def test_recompute_equilibrium(verbose=True, warnings=True, plot=True,
                               *args, **kwargs):
    ''' Test that spectral quantities recomputed under equilibrium assumption
    yields the same output as with non equilibrium routines when Tvib = Trot '''

    if plot:
        import matplotlib.pyplot as plt
        plt.ion()   # dont get stuck with Matplotlib if executing through pytest

    # Get spectrum
    s1 = load_spec(getTestFile('CO_Tgas1500K_mole_fraction0.01.spec'))
    s1.rescale_path_length(100)  # just for fun

    assert s1.is_at_equilibrium()
    s1.update('emisscoeff')

    # force non equilibrium calculation
    s2 = s1.copy()
    s2.conditions['thermal_equilibrium'] = False
    s2.update()
    assert 'emissivity_noslit' not in s2.get_vars()  # just a check update() was done at nonequilibrium

    # update s1 now (at equilibrium)
    s1.update()
    assert 'emissivity_noslit' in s1.get_vars()   # just a check update() was done at equilibrium

    s1.name = 'scaled with Kirchoff law'
    s2.name = 'scaled from emisscoeff + abscoeff with RTE'

    if verbose:
        print('Checked that scaling at equilibrium with Kirchoff law yields the ' +
              'same radiance as by solving the RTE from emisscoeff and abscoeff')

    # Now Compare both update processes
    assert s1.compare_with(s2, spectra_only='radiance_noslit', plot=plot)



def test_rescale_all_quantities(verbose=True, warnings=True, *args, **kwargs):

    new_mole_fraction = 0.5
    new_path_length = 0.1
    
    # Get spectrum
    s0 = load_spec(getTestFile('CO_Tgas1500K_mole_fraction0.01.spec'), binary=True)
    s0.update('all')     # start with all possible quantities in s0
    sscaled = s0.copy()
    sscaled.rescale_mole_fraction(new_mole_fraction)
    sscaled.rescale_path_length(new_path_length)
    s0.conditions['thermal_equilibrium'] = False   # to prevent rescaling with Kirchoff
    # remove emissivity_no_slit (speciifc to equilibrium)
    del s0._q['emissivity_noslit']
     
    
    # Determine all quantities that can be recomputed
    if verbose >= 2:
        import radis
        DEBUG_MODE = radis.DEBUG_MODE
        radis.DEBUG_MODE = True
    from radis.spectrum.rescale import get_reachable, ordered_keys, _build_update_graph
    # ordered_keys: all spectral quantities that can be rescaled
    can_be_recomputed = get_reachable(s0)
    # can_be_recomputed: all spectra quantities that can be rescaled for this 
    # particular spectrum
    update_paths = _build_update_graph(s0)
    # update_paths: which quantities are needed to recompute the others
    
    rescale_list = [k for k in ordered_keys if can_be_recomputed[k]]
    
    for quantity in rescale_list:
        all_paths = update_paths[quantity]
        if verbose:
            printm('{0} can be recomputed from {1}'.format(quantity, ' or '.join(
                    ['&'.join(combinations) for combinations in all_paths])))
        
        # Now let's test all paths
        for combinations in all_paths:
            if verbose:
                printm('> computing {0} from {1}'.format(quantity, '&'.join(combinations)))
            s = s0.copy()
            # Delete all other quantities
            for k in s.get_vars():
                if k not in combinations:
                    del s._q[k]
            
            s.update(quantity, verbose=verbose)
            
            # Now rescale
            s.rescale_mole_fraction(new_mole_fraction)
            s.rescale_path_length(new_path_length)
            
            # Compare
            assert s.compare_with(sscaled, spectra_only=quantity, plot=False)
            
    if verbose >= 2:
        radis.DEBUG_MODE = DEBUG_MODE
    
def _run_all_tests(verbose=True, warnings=True, *args, **kwargs):
    test_compression(verbose=verbose, warnings=warnings, *args, **kwargs)
    test_update_transmittance(verbose=verbose, warnings=warnings, *args, **kwargs)
    test_get_recompute(verbose=verbose, warnings=warnings, *args, **kwargs)
    test_recompute_equilibrium(verbose=verbose, warnings=warnings, *args, **kwargs)
    test_rescale_all_quantities(verbose=verbose, *args, **kwargs)

    return True


if __name__ == '__main__':
    print(('Testing test_rescale.py:', _run_all_tests(verbose=True)))



# -*- coding: utf-8 -*-
"""
Tools to convert spectroscopic constants with the 
conventions of Hubert and Herzberg (ω, Be, α, β, etc.)
to Dunham parameters (Ykl)

References
----------

https://en.wikipedia.org/wiki/Dunham_expansion

"""

from __future__ import print_function, absolute_import, division, unicode_literals

dunham2herzberg = {
        "Y01":(1, "Be"),
        "Y02":(-1, "De"),
        "Y03":(1, "He"),
        "Y04":(1, "Le"),
        "Y10":(1, "we"),
        "Y11":(-1, "alpha_e"),
        "Y12":(-1, "beta_e"),
        "Y20":(-1, "wexe"),
        "Y21":(-1, "gamma_e"),
        "Y30":(1, "weye"),
        "Y40":(1, "weze"),
        }
''' dict: {Yij: (sign, coeff) } 
conversion of Dunham spectroscopic coefficients to Herzberg convention '''

## Invert the dictionary.
herzberg2dunham = {v:(sign,k) for k,(sign,v) in dunham2herzberg.items()}
''' dict: {Yij: (sign, coeff)
conversion of Herberg convention to Dunham spectroscopic coefficients 

See Also
--------

:py:func:`~radis.db.utils.get_dunham_coefficients`,
:py:func:`~radis.db.utils.get_herzberg_coefficients` 
'''

# Name of all Herzberg coefficients
#herzberg_coefficients_elec = ['Te']
#'''list: Herzberg coefficients used to calculate pure electronic energies'''
herzberg_coefficients_vib = ['we', 'wexe', 'weye', 'weze', 'weae', 'webe']
'''list: Herzberg coefficients used to calculate pure vibrational energies'''
herzberg_coefficients_rot = ['Be', 'De', 'He', 'Le']
'''list: Herzberg coefficients used to calculate pure rotational energies'''
herzberg_coefficients_rovib = ['alpha_e', 'beta_e', 'gamma_e', 'delta_e', 'eta_e', 'pi_e']
'''list: Herzberg coefficients used in vibrational-rotational coupling terms for 
rovibrational energies'''

herzberg_coefficients = (herzberg_coefficients_vib  +
                         herzberg_coefficients_rot + herzberg_coefficients_rovib)
'''list: Herzberg coefficients used to calculate rovibrational energies in
:func:`~radis.levels.dunham.Fv`, :func:`~radis.levels.dunham.Gv`

See Also
--------

:py:func:`~radis.db.utils.get_herzberg_coefficients` 

'''


# Sanity check (test coefficients defined in conversion dictionaries are valid
# Dunham and Herzberg coefficients)
for k in dunham2herzberg:
    assert k.startswith('Y')
for k in herzberg2dunham:
    assert k in herzberg_coefficients

def get_convention(coefficients):
    ''' Returns if we're using the Herzberg or Dunham convention for 
    spectrosopic coefficients, and returns the associated coefficients
    
    Parameters
    ----------
    
    coefficients:
        list
    
    Returns
    -------
    
    convention: str
        'dunham' or 'herzberg'
        
    coefficients: dict
        list of coefficient names 
    '''
    
    from radis.misc.basics import partition
    from radis.db.utils import ignore_trailing_number
    
    assert len(coefficients) > 0
    
    herzberg_coeffs, non_herzberg_coeffs = partition(lambda x: ignore_trailing_number(x) in herzberg_coefficients,
                                                     coefficients)
    for k in non_herzberg_coeffs:
        if not k.startswith('Y'):
            raise ValueError('Unexpected Spectroscopic coefficient: {0}'.format(k))
    dunham_coeffs = non_herzberg_coeffs
    
    if len(dunham_coeffs) > 0 and len(herzberg_coeffs) > 0:
        raise ValueError('Both Dunham ({0}) and Herzberg ({1}) conventions used'.format(
                dunham_coeffs, herzberg_coeffs)+'. Choose one only')
        
    if len(dunham_coeffs) > 0:
        return 'dunham'
    else:
        return 'herzberg'

if __name__ == '__main__':
    
    assert get_convention(['wexe']) == 'herzberg'
    assert get_convention(['wexe1']) == 'herzberg'
    assert get_convention(['Y01', 'Y11']) == 'dunham'
    import pytest
    with pytest.raises(ValueError):
        # mixed conventions: should fail
        get_convention(['wexe', 'Y01', 'Y11'])
        


# -*- coding: utf-8 -*-
"""
Created on Tue Jul 18 17:59:45 2017

@author: erwan


Rovibrational Energy for Vibrating Rotor

Ex: used for CO2 X(1Σ+) state

References
----------





-----------------------------------------------------------------------------


"""

from __future__ import division, absolute_import, print_function, unicode_literals

from radis.levels.dunham import Gv, Fv
from radis.db.conventions import (herzberg_coefficients_rot, herzberg_coefficients_rovib,
                                  herzberg_coefficients_vib)

def EvJ_uncoupled_vibrating_rotor(v1, v2, l2, v3, J, coeff_dict, gv1=1, gv2=1, gv3=1, 
                                  remove_ZPE=True):
    ''' Rovibrational energy of an uncoupled vibrating rotor
    
    Parameters
    ----------

    v1: int
        vibrational state 

    v2: int
        vibrational state 

    l2: int
        vibrational state 

    v3: int
        vibrational state 

    J: int
        rotational state

    coeff_dict: dict
        dictionary of :py:data:`~radis.db.conventions.herzberg_coefficients`, 
        which can be numbered for the different vibration modes. Example::
            
            {'we1': 1333.93,
             'we2': 667.47,
             'we3': 2349.16,
             'wexe1': 2.93,
             'wexe2': -0.38,
             'wexe3': 12.47,
             'Be': 0.39022,
             'De': 1.333e-07,
             'He': 9e-15}
            
    gv1, gv2, gv3: int
        degeneracies of each vibrational mode. Default ``1, 1, 1``::
            
            1,2,1 for CO2
    
    remove_ZPE: boolean
        if ``True``, removes energy of ground state vibrational level (zero-point-energy)

    Returns
    -------

    E: float
        energy of state in cm-1 

    References
    ----------
    
    Klarenaar et al, "Time evolution of vibrational temperatures in a CO 2 glow 
    discharge measured with infrared absorption spectroscopy", doi 10.1088/1361-6595/aa902e,
    and the references there in.

    '''
    
    coeff_dict = coeff_dict.copy()
    
    coeffs_vib = {'1':{}, '2':{}, '3':{}}
    coeffs_rot = {}
    
    # Split coeffs between the different vibration and rotation modes:
    for k,v in coeff_dict.items():
        if k in herzberg_coefficients_rot:
            coeffs_rot[k] = v
        elif k[:-1] in herzberg_coefficients_vib:
            vib_mode = k[-1]
            coeffs_vib[vib_mode][k[:-1]] = v
        elif k in herzberg_coefficients_rovib:
            raise NotImplementedError('Mixed term {0} not implemented for uncoupled rovib model'.format(
                    k))
        else:
            raise KeyError('Unexpected coefficient: {0}'.format(k))
    
    coeffs_vib1 = coeffs_vib['1']
    coeffs_vib2 = coeffs_vib['2']
    coeffs_vib3 = coeffs_vib['3']
    
    # Get rotational coeffs:
    coeffs_rot = {k:v for (k,v) in coeff_dict.items() if k in herzberg_coefficients_rot}
    
    # Energies
    G1 = Gv(v1, gv=gv1, **coeffs_vib1)
    G2 = Gv(v2, gv=gv2, **coeffs_vib2)
    G3 = Gv(v3, gv=gv3, **coeffs_vib3)
    G = G1 + G2 + G3
    F = Fv(0, J, **coeffs_rot)      # Uncoupled model: v dependance ignored
    
    if remove_ZPE:
        ZPE = (Gv(0, **coeffs_vib1) + Gv(0, gv=2, **coeffs_vib2) + Gv(0, **coeffs_vib3))
    else:
        ZPE = 0

    return G + F - ZPE   # cm-1


def EvJah_uncoupled_vibrating_rotor(v1, v2, l2, v3, J, coeff_dict, remove_ZPE=True):
    ''' Return rovibrationalenergies for nu_1, nu_2, nu_3 vibrational modes
    Each energy is a tuple (E_harmonic, E_nonharmonic) to be
    used for instance in a Treanor distribution

    Defined for CO2 only. See:  Klarenaar et al. [1]_

    Energy of CO2 X(1Σg+) state with an independant harmonic oscillator
    approximation  (CO2-626 = first isotope)

    Note that this is less accurate than using CDSD energies in the (p, J, c, n)
    general denomination

    Parameters
    ----------

    v1: int
        vibrational state 

    v2: int
        vibrational state 

    l2: int
        vibrational state 

    v3: int
        vibrational state 

    J: int
        rotational state

    remove_ZPE: boolean
        if ``True``, removes energy of ground state vibrational level (zero-point-energy)

    Returns
    -------

    E: float
        energy of state in cm-1 

    References
    ----------
    
    .. [1] Eq.(10) in Klarenaar et al, "Time evolution of vibrational temperatures in a CO2 glow 
        discharge measured with infrared absorption spectroscopy", doi 10.1088/1361-6595/aa902e,
        and the references there in. 

    '''

    coeff_dict = coeff_dict.copy()

    we1 = coeff_dict.pop('we1')
    we2 = coeff_dict.pop('we2')
    we3 = coeff_dict.pop('we3')
    
    wexe1 = coeff_dict.pop('wexe1')
    wexe2 = coeff_dict.pop('wexe2')
    wexe3 = coeff_dict.pop('wexe3')
    
    Be = coeff_dict.pop('Be')
    De = coeff_dict.pop('De')
    He = coeff_dict.pop('He')
    
    if len(coeff_dict) > 0:
        raise NotImplementedError('Harmonic/anharmonic energy split not defined '+\
                                  'with the following spectroscopic constants: {0}'.format(
                                          list(coeff_dict.keys())))
    

    # Vibrational energy
    G1_h = we1*v1
    G1_a = - wexe1*v1*(v1-1)
    G1 = G1_h + G1_a

    G2_h = we2*v2
    G2_a = - wexe2*v2*(v2-1)
    G2 = G2_h + G2_a

    G3_h = we3*v3
    G3_a = - wexe3*v3*(v3-1)
    G3 = G3_h + G3_a

    G = G1 + G2 + G3

    # Rotational energy
    Bv = Be    # no coupling terms
    Dv = De
    Hv = He
    F = Bv*J*(J+1) - Dv*J**2*(J+1)**2 + Hv*J**3*(J+1)**3

    if remove_ZPE:
        ZPE = 0  # by construction here
    else:
        raise NotImplementedError

    return (G1_h, G1_a), (G2_h, G2_a), (G3_h, G3_a), F


#    return Te + G + F - ZPE   # cm-1


# %% Test

def _test(*args, **kwargs):
#    E_CO2X = E_CO2_626_X1SIGg
    from radis.db.utils import get_herzberg_coefficients
    coeffs = get_herzberg_coefficients('CO2', 1, 'X1SIGu+')
    E_CO2X = lambda v1, v2, l2, v3, J: EvJ_uncoupled_vibrating_rotor(v1, v2, l2, v3, J, coeff_dict=coeffs,
                                                                     gv1=1, gv2=2, gv3=1)
    print('CO2((0,0,0,0),J=0) -> CO2((0,0,0,1),J=0) energy: ',
          E_CO2X(0, 0, 0, 1, 0) - E_CO2X(0, 0, 0, 0, 0), 'cm-1')


if __name__ == '__main__':

    _test()



# -*- coding: utf-8 -*-
"""
Summary
-------

Models built around the :class:`~radis.spectrum.spectrum.Spectrum` class


Routine Listing
---------------

- :func:`~radis.spectrum.models.Transmittance`
- :func:`~radis.spectrum.models.Radiance`

- :func:`~radis.spectrum.models.calculated_spectrum`,
- :func:`~radis.spectrum.models.experimental_spectrum`,
- :func:`~radis.spectrum.models.transmittance_spectrum`,

-------------------------------------------------------------------------------


"""

from __future__ import print_function, absolute_import, division, unicode_literals
from radis.spectrum.spectrum import Spectrum
import numpy as np


# %% Array-to-Spectrum functions


def calculated_spectrum(w, I, wunit='nm', Iunit='mW/cm2/sr/nm',
                        conditions=None, cond_units=None, populations=None,
                        name=None): # -> Spectrum:
    ''' Convert ``(w, I)`` into a :py:class:`~radis.spectrum.spectrum.Spectrum`  
    object that has unit conversion, plotting and slit convolution capabilities


    Parameters    
    ----------

    w, I: np.array
        wavelength and intensity

    wunit: 'nm', 'cm-1'
        wavespace unit

    Iunit: str
        intensity unit (can be 'counts', 'mW/cm2/sr/nm', etc...). Default
        'mW/cm2/sr/nm' (note that non-convoluted Specair spectra are in 'mW/cm2/sr/µm')


    Other Parameters
    ----------------

    conditions: dict
        (optional) calculation conditions to be stored with Spectrum. Default ``None``

    cond_units: dict
        (optional) calculation conditions units. Default ``None``

    populations: dict
        populations to be stored in Spectrum. Default ``None``

    name: str
        (optional) give a name


    See Also
    --------

    :func:`~radis.spectrum.spectrum.transmittance_spectrum`, 
    :func:`~radis.spectrum.spectrum.experimental_spectrum`,
    :meth:`~radis.spectrum.spectrum.Spectrum.from_array`,
    :meth:`~radis.spectrum.spectrum.Spectrum.from_txt`,
    :func:`~radis.tools.database.load_spec`

    '''

    return Spectrum.from_array(np.array(w), np.array(I), 'radiance_noslit',
                               waveunit=wunit, unit=Iunit,
                               conditions=conditions, cond_units=cond_units,
                               populations=populations,
                               name=name)


def transmittance_spectrum(w, T, wunit='nm', Tunit='I/I0',
                           conditions=None, cond_units=None,
                           name=None): # -> Spectrum:
    ''' Convert ``(w, I)`` into a :py:class:`~radis.spectrum.spectrum.Spectrum`  
    object that has unit conversion, plotting and slit convolution capabilities


    Parameters    
    ----------

    w, I: np.array
        wavelength and transmittance (no slit)

    wunit: ``'nm'``, ``'cm-1'``
        wavespace unit

    Iunit: str
        intensity unit. Default ``'I/I0'``


    Other Parameters
    ----------------

    conditions: dict
        (optional) calculation conditions to be stored with Spectrum

    cond_units: dict
        (optional) calculation conditions units

    name: str
        (optional) give a name


    See Also
    --------

    :func:`~radis.spectrum.models.calculated_spectrum`, 
    :func:`~radis.spectrum.models.experimental_spectrum`,
    :meth:`~radis.spectrum.spectrum.Spectrum.from_array`,
    :meth:`~radis.spectrum.spectrum.Spectrum.from_txt`,
    :func:`~radis.tools.database.load_spec`

    '''

    return Spectrum.from_array(np.array(w), np.array(T), 'transmittance_noslit',
                               waveunit=wunit, unit=Tunit,
                               conditions=conditions, cond_units=cond_units,
                               name=name)


def experimental_spectrum(w, I, wunit='nm', Iunit='counts', medium='air',
                          conditions={}, cond_units=None, name=None): # -> Spectrum:
    ''' Convert ``(w, I)`` into a :py:class:`~radis.spectrum.spectrum.Spectrum` 
    object that has unit conversion and plotting
    capabilities. Convolution is not available as the spectrum is assumed to
    have be measured experimentally (hence it is already convolved with the slit function)


    Parameters    
    ----------

    w, I: np.array
        wavelength and intensity

    wunit: 'nm', 'cm-1'
        wavespace unit

    Iunit: str
        intensity unit (can be 'counts', 'mW/cm2/sr/nm', etc...). Default
        'counts' (default Winspec output)

    Other Parameters
    ----------------

    medium: 'air', 'vacuum'
        which medium. Default 'air'

    conditions: dict
        (optional) calculation conditions to be stored with Spectrum

    cond_units: dict
        (optional) calculation conditions units

    name: str
        (optional) give a name

    Examples
    --------
    
    Load and plot an experimental spectrum::
    
        from numpy import loadtxt
        from radis import experimental_spectrum
        w, I = loadtxt('my_file.txt').T    # assuming 2 columns 
        s = experimental_spectrum(w, I, Iunit='mW/cm2/sr/nm')
        s.plot()
        

    See Also
    --------

    :func:`~radis.spectrum.spectrum.calculated_spectrum`, 
    :func:`~radis.spectrum.spectrum.transmittance_spectrum`, 
    :meth:`~radis.spectrum.spectrum.Spectrum.from_array`,
    :meth:`~radis.spectrum.spectrum.Spectrum.from_txt`,
    :func:`~radis.tools.database.load_spec`

    '''
    
    if 'medium' in conditions:
        assert conditions['medium'] == medium
    else:
        conditions.update({'medium':medium})
        
    return Spectrum.from_array(np.array(w), np.array(I), 'radiance',
                               waveunit=wunit, unit=Iunit,
                               conditions=conditions, cond_units=cond_units,
                               name=name)


# -*- coding: utf-8 -*-
"""
Define warnings for radiation code, and how to deal with them

----------


"""

from __future__ import absolute_import, unicode_literals, print_function, division

import warnings
from radis.misc.printer import printr

# %% Spectrum warnings / errors
# -----------------------------
    
class SlitDispersionWarning(UserWarning):
    ''' Warning trigger if Slit dispersion is too large 
    '''

# %% Spectrum Factory warnings / errors
# -------------------------------------

# Errors

class EmptyDatabaseError(ValueError):
    pass

# Warnings

class GaussianBroadeningWarning(UserWarning):
    pass


class CollisionalBroadeningWarning(UserWarning):
    pass


class VoigtBroadeningWarning(UserWarning):
    pass


class MemoryUsageWarning(UserWarning):
    pass


class EmptyDatabaseWarning(UserWarning):
    ''' Trigger a warning if Line database is empty in the range considered '''
    pass

class OutOfRangeLinesWarning(UserWarning):
    ''' Trigger a warning if out of range neighbouring lines, that could have 
    an effect on the spectrume due to their broadening, cannot be found in the 
    database '''
    pass

class HighTemperatureWarning(UserWarning):
    ''' Warning triggered when the Line database seems inappropriate for the 
    temperatures considered
    '''
    pass


class NegativeEnergiesWarning(UserWarning):
    pass


class MissingSelfBroadeningWarning(UserWarning):
    ''' Self broadening is missing in Line Database. Usually, use Air broadening
    instead '''
    pass


class LinestrengthCutoffWarning(UserWarning):
    ''' Warning triggered when the cumulated linestrength after intensity cutoff
    has changed too much'''
    pass


class InputConditionsWarning(UserWarning):
    ''' Warning triggered when Spectrum input conditions are suspicious '''
    pass

class PerformanceWarning(UserWarning):
    ''' Warning triggered when it seems computation parameters are not optimized'''
    pass


WarningClasses = {'default': UserWarning,
                  'SlitDispersionWarning': SlitDispersionWarning,
                  'GaussianBroadeningWarning': GaussianBroadeningWarning,
                  'CollisionalBroadeningWarning': CollisionalBroadeningWarning,
                  'VoigtBroadeningWarning': VoigtBroadeningWarning,
                  'MemoryUsageWarning': MemoryUsageWarning,
                  'EmptyDatabaseWarning': EmptyDatabaseWarning,
                  'OutOfRangeLinesWarning': OutOfRangeLinesWarning,
                  'HighTemperatureWarning': HighTemperatureWarning,
                  'NegativeEnergiesWarning': NegativeEnergiesWarning,
                  'MissingSelfBroadeningWarning': MissingSelfBroadeningWarning,
                  'LinestrengthCutoffWarning': LinestrengthCutoffWarning,
                  'InputConditionsWarning': InputConditionsWarning,
                  'PerformanceWarning':PerformanceWarning,
                  }
''' dict: warnings used in RADIS Spectrum calculations.

You can selectively activate them by setting the warnings attribute of 
:class:`radis.lbl.factory.SpectrumFactory` 

See Also
--------

:py:data:`~radis.misc.warning.default_warning_status` 
'''

# Setup individual warnings. Value of keys can be:
# - 'warning' (default: just trigger a warning)
# - 'error' (raises an error on this warning)
# - 'ignore'  (do nothing)
# The key self.warnings['default'] will set the warning behavior for all
# other warnings
default_warning_status = {
    'default': 'warn',          # default
    'SlitDispersionWarning': 'warn',
    'GaussianBroadeningWarning': 'once',          # once per Spectrum calculation
    'CollisionalBroadeningWarning': 'once',       # once per Spectrum calculation
    'VoigtBroadeningWarning': 'once',             # once per Spectrum calculation
    # see also self.misc.warning_broadening_threshold for the treshold value
    'LinestrengthCutoffWarning': 'warn',
    'MemoryUsageWarning': 'warn',
    'EmptyDatabaseWarning': 'warn',
    'OutOfRangeLinesWarning': 'warn',
    'HighTemperatureWarning': 'warn',
    'NegativeEnergiesWarning': 'warn',    # warning if negative energies in database
    # warning if self-broadening abs coefficnet missing (Air is used instead)
    'MissingSelfBroadeningWarning': 'warn',
    'InputConditionsWarning': 'warn',
    'PerformanceWarning': 'warn',
}
''' dict: default status of warnings used in RADIS Spectrum calculations.

Value of keys can be:

- 'warning' (default: just trigger a warning)
- 'error' (raises an error on this warning)
- 'ignore'  (do nothing)

The key self.warnings['default'] will set the warning behavior for all
other warnings. All warnings can be disabled by setting the SpectrumFactory
:py:attr:`~radis.lbl.loader.DatabankLoader.warnings` attribute to ``False``.

See Also
--------

:py:data:`~radis.misc.warning.WarningClasses`, 
:py:func:`~radis.misc.warning.reset_warnings`

'''


def reset_warnings(status):
    ''' Reactivate warnings that are set 'once' per session in the Factory
    (unless all warnings have been set to False)

    Parameters
    ----------

    status: dict
        dictionary of Warnings with associated status

    '''
    
    if status == False:
        return

    for k, v in status.items():
        if v == 'once':
            WarningType = WarningClasses[k]
            warnings.simplefilter('default', WarningType)


def warn(message, category='default', status={}):
    ''' Trigger a warning, an error or just ignore based on the value defined
    in the :py:attr:`~radis.lbl.loader.DatabankLoader.warnings` dictionary

    The warnings can thus be deactivated selectively by setting the SpectrumFactory
    :attr:`~radis.lbl.loader.DatabankLoader.warnings` attribute. All warnings
    can be disabled by setting it to ``False``.

    Parameters
    ----------

    message: str
        what to print

    category: str
        one of the keys of self.warnings.

    status: dict
        status for all warning categories. Can be one of ``'warn'``, ``'ignore'``,
        ``'print'``, ``'error'``

    '''
    
    if status == False:
        return

    action = status[category]

    WarningType = WarningClasses[category]

    if action in 'warn':
        warnings.warn(WarningType(message))
    elif action == 'once':
        warnings.warn(WarningType(message))
        # keep 'once' but ignore WarningType with simplefilters
        warnings.simplefilter('ignore', WarningType)
    elif action == 'ignore':
        pass
    elif action == 'print':  # just print the message, in red
        printr(message)
    elif action == 'error':
        raise WarningType(message)
    else:
        raise ValueError('Unexpected action for warning: {0}'.format(action))


# Tests (on module load)



# ... test warnings are well defined
for k in default_warning_status.keys():
    assert k in WarningClasses

# ... and reciprocally, but they all have a default value
for k in WarningClasses.keys():
    assert k in default_warning_status



# -*- coding: utf-8 -*-
"""

Summary
-------

Classes to deal with multi-slabs configurations: 
    
- :func:`~radis.los.slabs.MergeSlabs` for several species at the same spatial position
- :func:`~radis.los.slabs.SerialSlabs` to add several spectra along the line-of-path

One Slab is just a :class:`~radis.spectrum.spectrum.Spectrum` object 


Examples
--------

See more examples in the 
`RADIS line-of-sight module <https://radis.readthedocs.io/en/latest/los/index.html>`__ 


-------------------------------------------------------------------------------

"""
# Todo:
#    
#- transport emisscoeff too 
# - emisscoeff default unit 


from __future__ import print_function, absolute_import, division, unicode_literals

from radis.spectrum.spectrum import Spectrum, is_spectrum
from radis.spectrum.utils import NON_CONVOLUTED_QUANTITIES
from radis.phys.convert import nm2cm
from radis.misc.basics import merge_lists
from radis.misc.utils import FileNotFoundError
from radis.misc.arrays import count_nans
from radis.misc.debug import printdbg
from warnings import warn
import numpy as np
from numpy import exp, arange, allclose, abs, diff

# %% Slabs / Multi-layers / Radiative Transfer Equation (RTE)
# ----------------------------------------------------------------------


def intersect(a, b):
    ''' Returns intersection of two dictionaries on values'''
    c = {}
    for k in set(a.keys()) & set(b.keys()):  # work in Python 2?
        c[k] = a[k] if (a[k] == b[k]) else 'N/A'
    return c


def SerialSlabs(*slabs, **kwargs):
    # type: (*Spectrum, **dict) -> Spectrum
    r''' Adds several slabs along the line-of-sight. 
    You can also use::
        
        s1>s2

    Serial spectrum ``1>2`` is calculated with Eqn (4.2) of the [RADIS-2018]_ article, 
    generalized to N slabs :
        
    .. math::
        I_{\lambda, 1>2} = I_{\lambda, 1} \tau_{\lambda, 2} + I_{\lambda, 2}
        
        \tau_{\lambda, 1+2} = \tau_{\lambda, 1} \cdot \tau_{\lambda, 2}
        
    where
    
        .. math:: I_{\lambda}, \tau_{\lambda}
        
    are the radiance and transmittance of the two slabs ``1`` and ``2``. 
    Radiance and transmittance are calculated if not given in the 
    initial slabs (if possible). 
    

    Parameters    
    ----------

    slabs: list of Spectra, each representing a slab
        line-of-sight::
                
            slabs       [0]     [1]  ............... [n]     
                         :       :                    :         \====
            light        *   ->  *        ->          *    ->    )===  observer
                                                                /====

    resample_wavespace: ``'never'``, ``'intersect'``, ``'full'``
        what to do when spectra have different wavespaces:
            
        - If ``'never'``, raises an error
        - If ``'intersect'``, uses the intersection of all ranges, and resample
          spectra on the most resolved wavespace. 
        - If ``'full``', uses the overlap of all ranges, resample spectra on the 
          most resolved wavespace, and fill missing data with 0 emission and 0
          absorption
        
        Default ``'never'``

    out: ``'transparent'``, ``'nan'``, ``'error'``
        what to do if resampling is out of bounds:
            
        - ``'transparent'``: fills with transparent medium. 
        - ``'nan'``: fills with nan. 
        - ``'error'``: raises an error. 
        
        Default ``'nan'``

    Other Parameters
    ----------------
    
    verbose: bool
        if ``True``, more blabla. Default ``False``

    Returns
    -------

    Spectrum object representing total emission and total transmittance as 
    observed at the output (slab[n+1]). Conditions and units are transported too,
    unless there is a mismatch then conditions are dropped (and units mismatch
    raises an error because it doesnt make sense)

    Examples
    --------

    Add s1 and s2 along the line of sight: s1 --> s2 ::

        s1 = calc_spectrum(...)
        s2 = calc_spectrum(...)
        s3 = SerialSlabs(s1, s2)
        
    The last line is equivalent to::
        
        s3 = s1>s2

    See Also
    --------

    :func:`~radis.los.slabs.MergeSlabs`

    See more examples in the :ref:`Line-of-Sight module <label_los_index>`
    
    '''
    # TODO: rewrite with 'recompute' list like in MergeSlabs ?
    
    if 'resample_wavespace' in kwargs:
        warn(DeprecationWarning("'resample_wavespace' replaced with 'resample'"))
        kwargs['resample'] = kwargs.pop('resample_wavespace')
    if 'out_of_bounds' in kwargs:
        warn(DeprecationWarning("'out_of_bounds' replaced with 'out'"))
        kwargs['out'] = kwargs.pop('out_of_bounds')

    # Check inputs, get defaults
    resample_wavespace = kwargs.pop('resample', 'never')   # default 'never'
    out_of_bounds = kwargs.pop('out', 'nan')               # default 'nan'    
    verbose = kwargs.pop('verbose', False)                   # type: bool
    if len(kwargs) > 0:
        raise ValueError('Unexpected input: {0}'.format(list(kwargs.keys())))

    if resample_wavespace not in ['never', 'intersect', 'full']:
        raise ValueError("resample should be one of: {0}".format
                         (', '.join(['never', 'intersect', 'full'])))

    if len(slabs) == 0:
        raise ValueError('Empty list of slabs')

    elif len(slabs) == 1:
        if not is_spectrum(slabs[0]):
            raise TypeError('SerialSlabs takes an unfolded list of Spectrum as ' +
                            'argument: *list (got {0})'.format(type(slabs[0])))
        return slabs[0]

    else:
        # recursively calculate serial slabs
        slabs = list(slabs)

#        # Check all items are Spectrum
        for s in slabs:
            _check_valid(s)

        # Recursively deal with the rest of Spectra --> call it s 
        sn = slabs.pop(-1)            # type: Spectrum
        s = SerialSlabs(*slabs, resample=resample_wavespace, out=out_of_bounds)

        # Now calculate sn and s in Serial
        quantities = {}
        unitsn = sn.units
        
        # make sure we use the same wavespace type (even if sn is in 'nm' and s in 'cm-1')
        # also make sure we use the same units
        waveunit = s.get_waveunit()

        # Make all our slabs copies with the same wavespace range
        # (note: wavespace range may be different for different quantities, but
        # equal for all slabs)
        s, sn = resample_slabs(waveunit, resample_wavespace, out_of_bounds,
                               s, sn)
        try:
            w = s._q['wavespace']
        except KeyError:
            raise KeyError('Cannot calculate the RTE if non convoluted quantities '+\
                           'are not defined. Got: {0}'.format(s.get_vars()))

        # Get all data
        # -------------
        
        I, In, T, Tn = None, None, None, None

        # To make it easier, the radiative transfer equation is solved with 'radiance_noslit' and 
        # 'transmittance_noslit' only. Here we first try to get these quantities: 

        # ... get sn quantities
        try:
            sn.update('transmittance_noslit', verbose=verbose)
        except ValueError:
            pass
        else:
            Tn = sn.get('transmittance_noslit', wunit=waveunit,
                           Iunit=unitsn['transmittance_noslit'])[1]
        try:
            sn.update('radiance_noslit', verbose=verbose)
        except ValueError:
            pass
        else:
            In = sn.get('radiance_noslit', wunit=waveunit,
                            Iunit=unitsn['radiance_noslit'])[1]
        # ... get s quantities
        try:
            s.update('transmittance_noslit', verbose=verbose)
        except ValueError:
            pass
        else:
            T = s.get('transmittance_noslit', wunit=waveunit,
                         Iunit=unitsn['transmittance_noslit'])[1]
        try:
            s.update('radiance_noslit', verbose=verbose)
        except ValueError:
            pass
        else:
            I = s.get('radiance_noslit', wunit=waveunit,
                         Iunit=unitsn['radiance_noslit'])[1]

        # Solve radiative transfer equation
        # ---------------------------------

        if I is not None and In is not None:
            # case where we may use SerialSlabs just to compute the products of all transmittances
            quantities['radiance_noslit'] = (w, I * Tn + In)

        if T is not None:  # note that we dont need the transmittance in the inner
                                         # slabs to calculate the total radiance
            quantities['transmittance_noslit'] = (w, Tn * T)

        # Get conditions (if they're different, fill with 'N/A')
        conditions = intersect(s.conditions, sn.conditions)
        conditions['waveunit'] = waveunit
        # sum path lengths
        if 'path_length' in s.conditions and 'path_length' in sn.conditions:
            conditions['path_length'] = s.conditions['path_length'] + sn.conditions['path_length']

        cond_units = intersect(s.cond_units, sn.cond_units)

        # name
        name = _serial_slab_names(s, sn)

        return Spectrum(quantities=quantities, conditions=conditions,
                        cond_units=cond_units, units=unitsn,
                        name=name)

def _serial_slab_names(s, sn):
    # type: (Spectrum, Spectrum) -> Spectrum
    name_s = s.get_name()
    name_sn = sn.get_name()
    if '//' in name_s and not '>>' in name_s:
        name_s = '({0})'.format(name_s)
    if '//' in name_sn:
        name_sn = '({0})'.format(name_sn)
    return '{0}>>{1}'.format(name_s, name_sn)


def _check_valid(s):
    # type: (Spectrum) -> bool
    ''' Check s is a valid Spectrum object. Raises an error if not 
    
    Valid if:
        
    - is a Spectrum
    
    Also print a warning if:
        
    - quantities used for solving the LOS have nan
    
    '''

    if not is_spectrum(s):
        raise TypeError('All inputs must be Spectrum objects (got: {0})'.format(
            type(s)))
    sdict = s._get_items()
    for k in sdict.keys():
        if (k in ['transmittance_noslit', 'radiance_noslit', 'abscoeff', 'emisscoeff'] 
                and count_nans(sdict.get(k)) > 0):
            warn('Nans detected in Spectrum object for multi-slab operation. ' +
                 'Results may be wrong!')

    return True


def _has_quantity(quantity, *slabs):
    # type: (str, *Spectrum) -> bool
    slabs = list(slabs)
    b = True
    for s in slabs:
        b *= quantity in slabs
    return b

def resample_slabs(waveunit, resample_wavespace, out_of_bounds='nan', *slabs):
    # type: (str, str, str, *Spectrum) -> *Spectrum
    ''' Resample slabs on the same wavespace: if the range are differents, 
    depending on the mode we may fill with optically thin media, or raise an
    error 


    Parameters    
    ----------

    waveunit: ``'nm'``, ``'cm-1'``
        which wavespace we're working in 

    resample_wavespace: 'never', 'intersect', 'full'
        what to do when spectra have different wavespaces:
            
        - If 'never', raises an error
        - If 'intersect', uses the intersection of all ranges, and resample
          spectra on the most resolved wavespace. 
        - If 'full', uses the overlap of all ranges, resample spectra on the 
          most resolved wavespace, and fill missing data with 0 emission and 0
          absorption
        
        Default 'never'

    out_of_bounds: 'transparent', 'nan', 'error'
        what to do if resampling is out of bounds:
            
        - 'transparent': fills with transparent medium. 
        - 'nan': fills with nan. 
        - 'error': raises an error. 
        
        Default ``'nan'``

    *slabs: list of Spectrum objects


    Returns
    -------

    slabs: list of Spectrum objects
        resampled copies of inputs Spectra. All now have the same wavespace

    '''

    # Check wavespace is the same
    def same_wavespace(wl):
        if not all([len(w) == len(wl[0]) for w in wl[1:]]):
            return False
        elif not all([allclose(w, wl[0]) for w in wl[1:]]):
            return False
        else:
            return True

    # Work on copies
    slabs = [s.copy() for s in slabs]

    # Get all keys
    keys = merge_lists([s.get_vars() for s in slabs])

    for k in keys:
        # We ensure all quantities are resampled. Note that .resample() deals
        # with all quantities so usually spectra are fully corrected after the
        # first iteration. However, it also deals with cases where quantities
        # have different wavespaces (ex: radiance and radiance_noslit are not
        # defined on the same range)
        slabsk = [s for s in slabs if k in s.get_vars()]  # note that these are
        # references to the actual Spectrum copy
        wl = [s.get(k, wunit=waveunit)[0] for s in slabsk]
        if not same_wavespace(wl):
            # resample slabs if allowed
            if resample_wavespace == 'never':
                raise ValueError('All wavelengths/wavenumbers must be the same for ' +
                                 'multi slabs configurations. Consider using '+\
                                 "`resample='intersect'` or " +\
                                 "`resample='full'`")
            elif resample_wavespace == 'full':
                # ... get bounds
                wmin = min([w.min() for w in wl])   # minimum of all
                wmax = max([w.max() for w in wl])   # maximum of all
                dw = min([abs(diff(w)).min() for w in wl])  # highest density
                wnew = arange(wmin, wmax+dw, dw)
                if wnew[-1] > wmax:  # sometimes arange doesnt work as expected
                    wnew = wnew[:-1]
                # ... copy the array of slabs not to modify the input
                # ... slabs themselves
                for s in slabsk:
                    s.resample(wnew, unit=waveunit, if_conflict_drop='convoluted',
                               out_of_bounds=out_of_bounds)
                # note: s.resample() fills with 0 when out of bounds
            elif resample_wavespace == 'intersect':
                # ... get bounds
                wmin = max([w.min() for w in wl])   # maximum of all
                wmax = min([w.max() for w in wl])   # minimum of all
                dw = min([abs(diff(w)).min() for w in wl])  # highest density
                wnew = arange(wmin, wmax+dw, dw)
                if wnew[-1] > wmax:  # sometimes arange doesnt work as expected
                    wnew = wnew[:-1]
                if len(wnew) == 0:
                    raise ValueError('Intersect range is empty')
                # ... copy the array of slabs not to modify the input
                # ... slabs themselves
                for s in slabsk:
                    s.resample(wnew, unit=waveunit, if_conflict_drop='convoluted',
                               out_of_bounds=out_of_bounds)
                # note: s.resample() fills with 0 when out of bounds
        # Now all our slabs have the same wavespace

    return slabs


def MergeSlabs(*slabs, **kwargs):
    # type: (*Spectrum, **dict) -> Spectrum
    ''' Combines several slabs into one. Useful to calculate multi-gas slabs. 
    Linear absorption coefficient is calculated as the sum of all linear absorption
    coefficients, and the RTE is recalculated to get the total radiance.
    You can also simply use::
        
        s1//s2 

    Merged spectrum ``1+2`` is calculated with Eqn (4.3) of the [RADIS-2018]_ article, 
    generalized to N slabs :
        
    .. math::
        
        j_{\lambda, 1+2} = j_{\lambda, 1} + j_{\lambda, 2}
        
        k_{\lambda, 1+2} = k_{\lambda, 1} + k_{\lambda, 2}
        
    where 
    
    .. math:: j_{\lambda}, k_{\lambda}
        
    are the emission coefficient and absorption coefficient of the two slabs ``1`` and ``2``. 
    Emission and absorption coefficients are calculated if not given in the 
    initial slabs (if possible). 
    
    Parameters    
    ----------

    slabs: list of Spectra, each representing a slab
        ``path_length`` must be given in Spectrum conditions, and equal for all 
        spectra.
        
        line-of-sight::
                
            slabs       
                        [0]        \====
            light       [1]  ->     )===  observer
                        [n]        /====

    Other Parameters
    ----------------

    kwargs input:

    resample: ``'never'``, ``'intersect'``, ``'full'``
        what to do when spectra have different wavespaces:
            
        - If ``'never'``, raises an error
        - If ``'intersect'``, uses the intersection of all ranges, and resample
          spectra on the most resolved wavespace. 
        - If ``'full'``, uses the overlap of all ranges, resample spectra on the 
          most resolved wavespace, and fill missing data with 0 emission and 0
          absorption
        
        Default ``'never'``

    out: ``'transparent'``, ``'nan'``, ``'error'``
        what to do if resampling is out of bounds:
            
        - ``'transparent'``: fills with transparent medium. 
        - ``'nan'``: fills with nan. 
        - ``'error'``: raises an error. 
        
        Default ``'nan'``

    optically_thin: boolean
        if ``True``, merge slabs in optically thin mode. Default ``False`` 

    verbose: boolean
        if ``True``, print messages and warnings. Default ``False``


    Returns
    -------

    Spectrum object representing total emission and total transmittance as 
    observed at the output. Conditions and units are transported too,
    unless there is a mismatch then conditions are dropped (and units mismatch
    raises an error because it doesnt make sense)


    Examples
    --------

    Merge two spectra calculated with different species (physically correct 
    only if broadening coefficients dont change much)::

        from radis import calc_spectrum, MergeSlabs
        s1 = calc_spectrum(...)
        s2 = calc_spectrum(...)
        s3 = MergeSlabs(s1, s2)

    The last line is equivalent to::
        
        s3 = s1//s2

    Load a spectrum precalculated on several partial spectral ranges, for a same 
    molecule (i.e, partial spectra are optically thin on the rest of the spectral 
    range)::

        from radis import load_spec, MergeSlabs
        spectra = []
        for f in ['spec1.spec', 'spec2.spec', ...]:
            spectra.append(load_spec(f))
        s = MergeSlabs(*spectra, resample='full', out='transparent')
        s.update()   # Generate missing spectral quantities
        s.plot()


    See Also
    --------

    :func:`~radis.los.slabs.SerialSlabs`
    
    See more examples in :ref:`Line-of-Sight module <label_los_index>`
    
    '''

    # Deprecation warnings
    if 'resample_wavespace' in kwargs:
        warn(DeprecationWarning("'resample_wavespace' replaced with 'resample'"))
        kwargs['resample'] = kwargs.pop('resample_wavespace')
    if 'out_of_bounds' in kwargs:
        warn(DeprecationWarning("'out_of_bounds' replaced with 'out'"))
        kwargs['out'] = kwargs.pop('out_of_bounds')

    # Check inputs, get defaults
    # inputs (Python 2 compatible)
    resample_wavespace = kwargs.pop('resample', 'never')   # default 'never'
    out_of_bounds = kwargs.pop('out', 'nan')               # default 'nan'
    optically_thin = kwargs.pop('optically_thin', False)             # default False
    verbose = kwargs.pop('verbose', False)             # type: bool
    debug = kwargs.pop('debug', False)                # type: bool
    if len(kwargs) > 0:
        raise ValueError('Unexpected input: {0}'.format(list(kwargs.keys())))

    # Check inputs
    if resample_wavespace not in ['never', 'intersect', 'full']:
        raise ValueError("'resample' should be one of: {0}".format
                         (', '.join(['never', 'intersect', 'full'])))

    if len(slabs) == 0:
        raise ValueError('Empty list of slabs')

    elif len(slabs) == 1:
        if not is_spectrum(slabs[0]):
            raise TypeError('MergeSlabs takes an unfolded list of Spectrum as ' +
                            'argument: (got {0})'.format(type(slabs[0])))
        return slabs[0]

    else:  # calculate serial slabs

        slabs = list(slabs)

#        # Check all items are valid Spectrum objects
        for s in slabs:
            _check_valid(s)

        # Check all path_lengths are defined and they exist
        try:
            path_lengths = [s.conditions['path_length'] for s in slabs]
        except KeyError:
            raise ValueError('path_length must be defined for all slabs in MergeSlabs. '+\
                             "Set it with `s.conditions['path_length']=`. ")
        if not all([L == path_lengths[0] for L in path_lengths[1:]]):
            raise ValueError('path_length must be equal for all MergeSlabs inputs' +
                             '  (got {0})'.format(path_lengths))

        # make sure we use the same wavespace type (even if sn is in 'nm' and s in 'cm-1')
        waveunit = slabs[0].get_waveunit()
        # Make all our slabs copies with the same wavespace range
        # (note: wavespace range may be different for different quantities, but
        # equal for all slabs)
        slabs = resample_slabs(waveunit, resample_wavespace, out_of_bounds, *slabs)
        w_noconv = slabs[0]._get_wavespace()

        # %%

        # Get conditions
        conditions = slabs[0].conditions
        conditions['waveunit'] = waveunit
        cond_units = slabs[0].cond_units
        units0 = slabs[0].units
        # Define conditions as intersection of everything (N/A if unknown)
        for s in slabs[1:]:
            conditions = intersect(conditions, s.conditions)
            cond_units = intersect(cond_units, s.cond_units)
            # units = intersect(units0, s.units)  # we're actually using [slabs0].units insteads

        # %% Get quantities that should be calculated
        # Try to keep all the quantities of the initial slabs:
        requested = merge_lists([s.get_vars() for s in slabs])
        recompute = requested[:]  # copy
        if ('radiance_noslit' in requested and not optically_thin):
            recompute.append('emisscoeff')
            recompute.append('abscoeff')
        if 'abscoeff' in recompute and 'path_length' in conditions:
            recompute.append('absorbance')
            recompute.append('transmittance_noslit')
                
        # To make it easier, we start from abscoeff and emisscoeff of all slabs
        # Let's recompute them all
        # TODO: if that changes the initial Spectra, maybe we should just work on copies
        for s in slabs:
            if 'abscoeff' in recompute and not 'abscoeff' in list(s._q.keys()):
                s.update('abscoeff', verbose=False)
                # that may crash if Spectrum doesnt have the correct inputs.
                # let update() handle that
            if 'emisscoeff' in recompute and not 'emisscoeff' in list(s._q.keys()):
                s.update('emisscoeff', verbose=False)
                # same

        # %% Calculate total emisscoeff and abscoeff
        added = {}

        # ... absorption coefficient (cm-1)
        if 'abscoeff' in recompute:
            # TODO: deal with all cases
            if __debug__:
                printdbg('... merge: calculating abscoeff k=sum(k_i)')
            abscoeff_eq = np.sum([s.get('abscoeff', wunit=waveunit, Iunit=units0['abscoeff'])[
                                 1] for s in slabs], axis=0)
            assert len(w_noconv) == len(abscoeff_eq)
            added['abscoeff'] = (w_noconv, abscoeff_eq)

        # ... emission coefficient
        if 'emisscoeff' in recompute:
            if __debug__:
                printdbg('... merge: calculating emisscoeff j=sum(j_i)')
            emisscoeff_eq = np.sum([s.get('emisscoeff', wunit=waveunit, Iunit=units0['emisscoeff'])[
                                 1] for s in slabs], axis=0)
            assert len(w_noconv) == len(emisscoeff_eq)
            added['emisscoeff'] = (w_noconv, emisscoeff_eq)
    
        # name
        name = '//'.join([s.get_name() for s in slabs])
        
        # TODO: check units are consistent in all slabs inputs
        s = Spectrum(quantities=added, conditions=conditions,
                        cond_units=cond_units, units=units0,
                        name=name)
        
        # %% Calculate all quantities from emisscoeff and abscoeff
        
        if 'emissivity_noslit' in requested and (
                'thermal_equilibrium' not in s.conditions or s.is_at_equilibrium() != True):
            requested.remove('emissivity_noslit')
            if __debug__:
                printdbg('... merge: all slabs are not proven to be at equilibrium. '+\
                         'Emissivity was not calculated')
                
        # Add the rest of the spectral quantities afterwards:
        s.update([k for k in requested if k not in ['emisscoeff', 'abscoeff']], 
                  optically_thin=optically_thin, verbose=verbose)

        return s


# %% Tests


if __name__ == '__main__':
    from radis.test.los.test_slabs import _run_testcases
    print('Testing merge slabs: ', _run_testcases(verbose=True))



# -*- coding: utf-8 -*-
"""
Created on Mon May  7 17:34:52 2018

@author: erwan
"""

from __future__ import absolute_import, unicode_literals, division, print_function
from radis.lbl import SpectrumFactory
from radis.misc.utils import DatabankNotFound
from radis.misc.printer import printm
from radis.test.utils import IgnoreMissingDatabase, setup_test_line_databases
from radis.phys.convert import cm2nm
import pytest
import numpy as np
from radis.misc.progress_bar import ProgressBar
from radis import get_residual, sPlanck
import radis
import matplotlib.pyplot as plt

@pytest.mark.fast
def test_populations(plot=True, verbose=True, warnings=True, *args, **kwargs):
    ''' See wavelength difference in air and vacuum '''

    if plot:  # Make sure matplotlib is interactive so that test are not stuck in pytest
        plt.ion()

    try:
        if verbose:
            printm('>>> _test_media_line_shift')

        setup_test_line_databases()  # add HITRAN-CO-TEST in ~/.radis if not there

        sf = SpectrumFactory(wavelength_min=4500, wavelength_max=4600,
                             wstep=0.001,
                             parallel=False, bplot=False, cutoff=1e-30,
                             path_length=0.1, mole_fraction=400e-6,
                             isotope=[1], db_use_cached=True,
                             medium='vacuum',
                             broadening_max_width=10,
                             export_populations='rovib',
                             verbose=verbose)
        sf.warnings['MissingSelfBroadeningWarning'] = 'ignore'
        sf.load_databank('HITRAN-CO-TEST')
        
        # Populations cannot be calculated at equilibrium (no access to energy levels)
        s = sf.eq_spectrum(300)
        with pytest.raises(ValueError):  # .. we expect this error here!
            pops = s.get_populations('CO', isotope=1, electronic_state='X')

        # Calculate populations using the non-equilibrium module:
        s = sf.non_eq_spectrum(300, 300)
        pops = s.get_populations('CO', isotope=1, electronic_state='X')
        if not 'n' in list(pops['rovib'].keys()):
            raise ValueError('Populations not calculated')
        if plot:
            s.plot_populations()
            
        # Compare with factory
        # Plot populations:
        with pytest.raises(ValueError):
            sf.plot_populations()    # no isotope given: error expected
        if plot:
            sf.plot_populations('rovib', isotope=1)
            plt.close()  # no need to keep it open, we just tested the function

            
        # Test calculated quantities are there
        assert hasattr(sf.df1, 'Qref')
        assert hasattr(sf.df1, 'Qvib')
        assert hasattr(sf.df1, 'Qrotu')
        assert hasattr(sf.df1, 'Qrotl')

        # Test hardcoded populations
        assert np.isclose(pops['rovib']['n'].iloc[0], 0.0091853446840826653)
        assert np.isclose(pops['rovib']['n'].iloc[1], 0.027052543988733215)
        assert np.isclose(pops['rovib']['n'].iloc[2], 0.04345502115897712)
        
        return True

    except DatabankNotFound as err:
        assert IgnoreMissingDatabase(err, __file__, warnings)

#@pytest.mark.needs_connection
def test_optically_thick_limit_1iso(verbose=True, plot=True, *args, **kwargs):
    ''' Test that we find Planck in the optically thick limit 
    
    In particular, this test will fail if :
        
    - linestrength are not properly calculated
    
    - at noneq, linestrength and emission integrals are mixed up
    
    The test should be run for 1 and several isotopes, because different
    calculations paths are used internally, and this can lead to different
    errors.
    
    Also, this test is used to run with DEBUG_MODE = True, which will 
    check that isotopes and molecule ids are what we expect in all the 
    groupby() loops that make the production code very fast. 
    
    Notes
    -----
    
    switched from large band calculation with [HITRAN-2016]_ to a calculation with 
    the embedded [HITEMP-2010]_ fragment (shorter range, but no need to download files)
    
    '''
    
    if plot:  # Make sure matplotlib is interactive so that test are not stuck in pytest
        plt.ion()

    # Force DEBUG_MODE
    DEBUG_MODE = radis.DEBUG_MODE
    radis.DEBUG_MODE = True
    
    try:
        
#        wavelength_min = 4500
#        wavelength_max = 4800
        wavelength_min = cm2nm(2284.6)
        wavelength_max = cm2nm(2284.2)
    
        P = 0.017     # bar
        wstep = 0.001     # cm-1
            
        Tgas = 1200
        
        # %% Generate some CO2 emission spectra
        # --------------
        sf = SpectrumFactory(wavelength_min=wavelength_min, wavelength_max=wavelength_max,
                             molecule='CO2',
                             mole_fraction=1, path_length=0.05, cutoff=1e-25,
                             broadening_max_width=1,
                             export_populations = False, #'vib',
                             export_lines = False,
                             isotope=1,
                             use_cached=True,
                             wstep=wstep,
                             pseudo_continuum_threshold=0,
                             pressure=P,
                             verbose=False)
#        sf.fetch_databank('astroquery')
        sf.warnings['NegativeEnergiesWarning'] = 'ignore'
        sf.load_databank('HITEMP-CO2-TEST')
        pb = ProgressBar(3, active=verbose)
        s_eq = sf.eq_spectrum(Tgas=Tgas,  mole_fraction=1, 
                              name='Equilibrium')
        pb.update(1)
        s_2T = sf.non_eq_spectrum(Tvib=Tgas, Trot=Tgas, mole_fraction=1, 
                                  name='Noneq (2T)')
        pb.update(2)
        s_4T = sf.non_eq_spectrum(Tvib=(Tgas, Tgas, Tgas), Trot=Tgas, mole_fraction=1, 
                                  name='Noneq (4T)')
        pb.update(3)
        s_plck = sPlanck(wavelength_min=2000, #=wavelength_min, 
                         wavelength_max=5000, #=wavelength_max - wstep,   # there is a border effect on last point
                         T=Tgas)
        pb.done()
        
        # %% Post process: 
        # MAke optically thick, and compare with Planck
        
        for s in [s_eq, s_2T, s_4T]:
        
            s.rescale_path_length(1e6)
            
            if plot:
                
                nfig = 'test_opt_thick_limit_1iso {0}'.format(s.name)
                plt.figure(nfig).clear()
                s.plot(wunit='nm', nfig=nfig, lw=4)
                s_plck.plot(wunit='nm', nfig=nfig, Iunit='mW/cm2/sr/nm', lw=2)
                plt.legend()
                
            if verbose:
                printm("Residual between opt. thick CO2 spectrum ({0}) and Planck: {1:.2g}".format(
                        s.name, get_residual(s, s_plck, 'radiance_noslit', ignore_nan=True)))
            
#            assert get_residual(s, s_plck, 'radiance_noslit', ignore_nan=True) < 1e-3
            assert get_residual(s, s_plck, 'radiance_noslit', ignore_nan=True) < 2e-5
            
        if verbose:
            printm('Tested optically thick limit is Planck (1 isotope): OK')
    
    finally:
        # Reset DEBUG_MODE
        radis.DEBUG_MODE = DEBUG_MODE

#@pytest.mark.needs_connection
#@pytest.mark.needs_db_HITEMP_CO2_DUNHAM
#@pytest.mark.needs_connection
def test_optically_thick_limit_2iso(verbose=True, plot=True, *args, **kwargs):
    ''' Test that we find Planck in the optically thick limit 
    
    In particular, this test will fail if :
        
    - linestrength are not properly calculated
    
    - at noneq, linestrength and emission integrals are mixed up
    
    The test should be run for 1 and several isotopes, because different
    calculations paths are used internally, and this can lead to different
    errors.
    
    Also, this test is used to run with DEBUG_MODE = True, which will 
    check that isotopes and molecule ids are what we expect in all the 
    groupby() loops that make the production code very fast. 
    
    Notes
    -----
    
    switched from large band calculation with [HITRAN-2016]_ to a calculation with 
    the embedded [HITEMP-2010]_ fragment (shorter range, but no need to download files)
    
    '''
    
    if plot:  # Make sure matplotlib is interactive so that test are not stuck in pytest
        plt.ion()

    # Force DEBUG_MODE
    DEBUG_MODE = radis.DEBUG_MODE
    radis.DEBUG_MODE = True
    
    try:
        
#        wavelength_min = 4500
#        wavelength_max = 4800
        wavelength_min = cm2nm(2284.6)
        wavelength_max = cm2nm(2284.2)
#    2284.2, 2284.6,
#                             wstep=0.001,                # cm-1
        P = 0.017     # bar
        wstep = 0.001     # cm-1
        
            
        Tgas = 1200
        
        # %% Generate some CO2 emission spectra
        # --------------
        sf = SpectrumFactory(wavelength_min=wavelength_min, wavelength_max=wavelength_max,
                             mole_fraction=1, path_length=0.05, cutoff=1e-25,
                             broadening_max_width=1,
                             export_populations = False, #'vib',
                             export_lines = False,
                             molecule='CO2',
                             isotope=[1,2],
                             db_use_cached=True,
                             wstep=wstep,
                             pseudo_continuum_threshold=0,
                             pressure=P,
                             verbose=False)
#        sf.fetch_databank('astroquery')
        sf.load_databank('HITEMP-CO2-TEST')
        sf.warnings['NegativeEnergiesWarning'] = 'ignore'
#        sf.warnings['MissingSelfBroadeningWarning'] = 'ignore'
        pb = ProgressBar(3, active=verbose)
        s_eq = sf.eq_spectrum(Tgas=Tgas,  mole_fraction=1, 
                              name='Equilibrium')
        pb.update(1)
        s_2T = sf.non_eq_spectrum(Tvib=Tgas, Trot=Tgas, mole_fraction=1, 
                                  name='Noneq (2T)')
        pb.update(2)
        s_4T = sf.non_eq_spectrum(Tvib=(Tgas, Tgas, Tgas), Trot=Tgas, mole_fraction=1, 
                                  name='Noneq (4T)')
        pb.update(3)
        s_plck = sPlanck(wavelength_min=2000, #=wavelength_min, 
                         wavelength_max=5000, #=wavelength_max - wstep,   # there is a border effect on last point
                         T=Tgas)
        pb.done()
        
        # %% Post process: 
        # MAke optically thick, and compare with Planck
        
        for s in [s_eq, s_2T, s_4T]:
        
            s.rescale_path_length(1e6)
            
            if plot:
                
                nfig = 'test_opt_thick_limit_2iso {0}'.format(s.name)
                plt.figure(nfig).clear()
                s.plot(wunit='nm', nfig=nfig, lw=4)
                s_plck.plot(wunit='nm', nfig=nfig, Iunit='mW/cm2/sr/nm', lw=2)
                plt.legend()
                
            if verbose:
                printm("Residual between opt. thick CO2 spectrum ({0}) and Planck: {1:.2g}".format(
                        s.name, get_residual(s, s_plck, 'radiance_noslit', ignore_nan=True)))
            
#            assert get_residual(s, s_plck, 'radiance_noslit', ignore_nan=True) < 1e-3
            assert get_residual(s, s_plck, 'radiance_noslit', ignore_nan=True) < 2e-5
            
        if verbose:
            printm('Tested optically thick limit is Planck (2 isotopes): OK')
    
    finally:
        # Reset DEBUG_MODE
        radis.DEBUG_MODE = DEBUG_MODE


def _run_testcases(verbose=True, plot=True):
    
    test_populations(plot=plot, verbose=verbose)
    test_optically_thick_limit_1iso(plot=plot, verbose=verbose)
    test_optically_thick_limit_2iso(plot=plot, verbose=verbose)

if __name__ == '__main__':
    _run_testcases()



# -*- coding: utf-8 -*-
"""
# NeQ - Non Equilibrium Plasma Database and Processing Tools 

See Readme for more information 

"""

from __future__ import absolute_import, division, print_function, unicode_literals



# -*- coding: utf-8 -*-
"""

Examples
--------

Run all tests:

>>> pytest       (in command line, in project folder)

Run only fast tests 

>>> pytest -m fast

"""

from __future__ import absolute_import, unicode_literals, division, print_function
from radis.spectrum.models import calculated_spectrum
from radis.phys.convert import nm2cm
from radis.misc.utils import DatabankNotFound
import numpy as np
from numpy import allclose, linspace
import matplotlib.pyplot as plt
from radis.test.utils import IgnoreMissingDatabase, setup_test_line_databases
from radis.misc.printer import printm
from os.path import basename
import pytest
from warnings import filterwarnings, catch_warnings

fig_prefix = basename(__file__)+': '

# %% Test routines


@pytest.mark.fast
def test_populations(verbose=True, *args, **kwargs):
    ''' Test that vib and rovib populations are calculated correctly '''

    from radis.lbl import SpectrumFactory
    from radis.misc.basics import all_in

    export = ['vib', 'rovib']
    sf = SpectrumFactory(2000, 2300,
                         export_populations=export,
                         db_use_cached=True,
                         cutoff=1e-25,
                         isotope='1')
    sf.warnings.update({'MissingSelfBroadeningWarning': 'ignore',
                        'VoigtBroadeningWarning': 'ignore'})
    sf.load_databank('HITRAN-CO-TEST')

    s = sf.non_eq_spectrum(2000, 2000)

    pops = sf.get_populations(export)
    if not all_in(['rovib', 'vib'], list(pops['CO'][1]['X'].keys())):
        raise AssertionError(
            'vib and rovib levels should be defined after non_eq_spectrum calculation!')
    if not 'nvib' in list(pops['CO'][1]['X']['vib'].keys()):
        raise AssertionError(
            'Vib populations should be defined after non_eq_spectrum calculation!')

    s = sf.eq_spectrum(300)

    pops = sf.get_populations(export)
    if 'nvib' in list(pops['CO'][1]['X']['vib'].keys()):
        raise AssertionError(
            'Vib levels should not be defined anymore after eq_spectrum calculation!')

    # Any of these is True and something went wrong
    s = sf.non_eq_spectrum(2000, 2000)
    s2 = sf.non_eq_spectrum(300, 300)

    #printm(all(s2.get_vib_levels(isotope=1) == s.get_vib_levels(isotope=1)))
    assert not (s2.get_vib_levels() is s.get_vib_levels())
    assert not (s2.get_rovib_levels() == s.get_rovib_levels()).all().all()
    assert not (s2.get_rovib_levels() is s.get_rovib_levels())

    return True  # if no AssertionError


@pytest.mark.fast
def test_rescaling_path_length(debug=False, plot=False, verbose=True, warnings=True,
                               *args, **kwargs):
    ''' Test rescaling functions '''

    if plot:  # Make sure matplotlib is interactive so that test are not stuck
        plt.ion()

    try:
        from radis.lbl import SpectrumFactory

        setup_test_line_databases()  # add HITRAN-CO-TEST in ~/.radis if not there

        Tgas = 1500
        sf = SpectrumFactory(
            wavelength_min=4400,
            wavelength_max=4800,
            mole_fraction=0.01,
            #                         path_length=0.1,
            cutoff=1e-25,
            wstep=0.005,
            isotope=[1],
            db_use_cached=True,
            self_absorption=True,
            verbose=verbose)
        sf.warnings['MissingSelfBroadeningWarning'] = 'ignore'
#        sf.warnings['NegativeEnergiesWarning'] = 'ignore'
        sf.load_databank('HITRAN-CO-TEST')
        s1 = sf.non_eq_spectrum(Tgas, Tgas, path_length=0.01)
        s2 = sf.non_eq_spectrum(Tgas, Tgas, path_length=3)
        s1.rescale_path_length(3)

        if plot:
            fig = plt.figure(fig_prefix+'Rescaling path length')
            s2.plot('radiance_noslit', nfig=fig.number, lw=3, label='L=3m')
            s1.plot('radiance_noslit', nfig=fig.number,
                    color='r', label='L=0.01m, rescaled to 3m')
            plt.title('Non optically thin rescaling')
            plt.legend()
            plt.tight_layout()

        if verbose:
            printm('Test rescaling:')
            printm('... Difference: {0:.2f}%'.format(
                abs(s1.get_power()/s2.get_power()-1)*100))

        assert np.isclose(s2.get_power(), s1.get_power(), 2e-3)

    except DatabankNotFound as err:
        assert IgnoreMissingDatabase(err, __file__, warnings)


@pytest.mark.fast
def test_rescaling_mole_fraction(debug=False, plot=False, verbose=True, warnings=True,
                                 *args, **kwargs):
    ''' Test rescaling functions '''

    from radis.lbl import SpectrumFactory

    if plot:  # Make sure matplotlib is interactive so that test are not stuck
        plt.ion()

    try:
        setup_test_line_databases()  # add HITRAN-CO-TEST in ~/.radis if not there

        Tgas = 1500
        sf = SpectrumFactory(
            wavelength_min=4400,
            wavelength_max=4800,
            #                     mole_fraction=1,
            path_length=0.1,
            mole_fraction=0.01,
            cutoff=1e-25,
            wstep=0.005,
            isotope=[1],
            db_use_cached=True,
            self_absorption=True,
            verbose=verbose)
        sf.warnings['MissingSelfBroadeningWarning'] = 'ignore'
        sf.warnings['NegativeEnergiesWarning'] = 'ignore'
        sf.load_databank('HITRAN-CO-TEST')
        error = []
        N = [1e-3, 1e-2, 1e-1, 0.3, 0.6, 1]  # first is ref
        for Ni in N:
            s1 = sf.non_eq_spectrum(Tgas, Tgas, mole_fraction=N[0])
            sN = sf.non_eq_spectrum(Tgas, Tgas, mole_fraction=Ni)
            s1.rescale_mole_fraction(Ni)
            error.append(sN.get_power() / s1.get_power())

        if plot:
            plt.figure(fig_prefix+'Rescaling mole fractions')
            plt.plot(N, error, '-ok')
            plt.scatter(N[0], error[0], s=200, facecolors='none',
                        edgecolors='r', label='reference')
            plt.xlabel('Mole fraction')
            plt.ylabel('scaled energy / ab initio energy')
            plt.xscale('log')
            plt.legend()
            plt.title('Effect of scaling mole fraction w/o lineshape update')
            plt.tight_layout()

        # less than 1% error when rescaling from 1e-3 to 0.6
        assert abs(error[-2]-1) < 0.01

    except DatabankNotFound as err:
        assert IgnoreMissingDatabase(err, __file__, warnings)


@pytest.mark.fast
def test_medium(plot=False, verbose=True, debug=False, warnings=True, *args, **kwargs):
    ''' Test effect of propagating medium '''

    from radis.lbl.factory import SpectrumFactory

    if plot:  # Make sure matplotlib is interactive so that test are not stuck
        plt.ion()

    T = 300

    try:
        setup_test_line_databases()  # add HITRAN-CO-TEST in ~/.radis if not there

        pl = SpectrumFactory(
            wavenum_min=2171.5,
            wavenum_max=2174,
            mole_fraction=0.01,
            medium='vacuum',
            isotope='1,2')
        pl.warnings['MissingSelfBroadeningWarning'] = 'ignore'
        pl.load_databank('HITRAN-CO-TEST')
        s = pl.non_eq_spectrum(Tvib=T, Trot=T)  # , Ttrans=300)

        pla = SpectrumFactory(
            wavenum_min=2171.5,
            wavenum_max=2174,
            mole_fraction=0.01,
            medium='air',
            isotope='1,2')
        pla.load_databank('HITRAN-CO-TEST')
        s_air = pla.non_eq_spectrum(Tvib=T, Trot=T)  # , Ttrans=300)

        if plot:
            plt.figure(fig_prefix+'Propagating medium conversions')
            s.plot(nfig='same', lw=3, medium='vacuum', label='vacuum')
            s.plot(nfig='same', medium='air', label='air')

        assert np.allclose(s.get_wavenumber(), s_air.get_wavenumber())
        assert np.allclose(s.get_wavelength(medium='vacuum'),
                           s_air.get_wavelength(medium='vacuum'))
        assert np.allclose(s.get_wavelength(medium='air'),
                           s_air.get_wavelength(medium='air'))
        assert all(s.get_wavelength(medium='vacuum')
                   > s.get_wavelength(medium='air'))

    except DatabankNotFound as err:
        assert IgnoreMissingDatabase(err, __file__, warnings)


def _run_testcases(plot=False, verbose=True, debug=False, warnings=True, *args, **kwargs):
    ''' Test procedures

    Parameters
    ----------

    debug: boolean
        swamps the console namespace with local variables. Default ``False``

    '''

    # Test populations
    # ----------
    test_populations(verbose=verbose, *args, **kwargs)

    # Test conversion of intensity cm-1 works
    # -------------
#    test_intensity_conversion(debug=debug, verbose=verbose, *args, **kwargs)   # Moved in RADIS

    # Test updating / rescaling functions (no self absorption)
    # ---------
#    test_rescaling_function(debug=debug, *args, **kwargs)     # Moved in RADIS
    test_rescaling_path_length(plot=plot, verbose=verbose, debug=debug,
                               warnings=warnings, *args, **kwargs)
    test_rescaling_mole_fraction(plot=plot, verbose=verbose, debug=debug,
                                 warnings=warnings, *args, **kwargs)

    # Test propagating medium
    test_medium(plot=plot, verbose=verbose, debug=debug,
                warnings=warnings, *args, **kwargs)

    return True


if __name__ == '__main__':
    printm('Test spectrum: ', _run_testcases(debug=False, plot=True))



# -*- coding: utf-8 -*-
"""
Test query functions

-------------------------------------------------------------------------------

"""

from __future__ import print_function, absolute_import, division, unicode_literals
import pytest
from radis.io.query import fetch_astroquery
from radis.io.query import Hitran, CACHE_FILE_NAME
from radis.misc.cache_files import DeprecatedFileError
from os.path import exists, join


# ignored by pytest with argument -m "not needs_connection"
@pytest.mark.needs_connection
def test_fetch_astroquery(verbose=True, *args, **kwargs):
    ''' Test astroquery '''
    df = fetch_astroquery('CO2', 1, 2200, 2400, verbose=verbose,
                          cache=False)

    assert df.iloc[0].id == 2
    assert df.iloc[0].iso == 1

    return


# ignored by pytest with argument -m "not needs_connection"
@pytest.mark.needs_connection
def test_fetch_astroquery_empty(verbose=True, *args, **kwargs):
    ''' Test astroquery: get a spectral range where there are no lines'''
    df = fetch_astroquery(2, 1, 25000, 50000, verbose=verbose,
                          cache=False)  # 200-400 nm

    assert len(df) == 0

    return


# ignored by pytest with argument -m "not needs_connection"
@pytest.mark.needs_connection
def test_fetch_astroquery_cache(verbose=True, *args, **kwargs):
    ''' Test astroquery cache file generation'''
    
    df = fetch_astroquery('CO2', 1, 2200, 2400, verbose=verbose,
                          cache='regen', # Delete existing cache file if needed
                          metadata={'_test':True})

    # assert cache file was created
    assert exists(join(Hitran.cache_location, CACHE_FILE_NAME.format(**{'molecule':'CO2',
                                                                        'isotope':1,
                                                                        'wmin':2200,
                                                                        'wmax':2400})))
    
    # Try to load with different metadata: expect a DeprecatedFileError
    with pytest.raises(DeprecatedFileError):
        df2 = fetch_astroquery('CO2', 1, 2200, 2400, verbose=verbose,
                              cache='force',  # force load of cache file
                              metadata={})
    # Try to load with correct metadata. Expect to work and return the dataframe
    df2 = fetch_astroquery('CO2', 1, 2200, 2400, verbose=verbose,
                          cache='force',  # force load of cache file
                          metadata={'_test':True})
    assert (df == df2).all().all()
    
    return



def _run_testcases(verbose=True, *args, **kwargs):

    test_fetch_astroquery(verbose=verbose, *args, **kwargs)
    test_fetch_astroquery_empty(verbose=verbose, *args, **kwargs)
    test_fetch_astroquery_cache(verbose=verbose, *args, **kwargs)

    return True


if __name__ == '__main__':
    print('test_query.py: ', _run_testcases(verbose=True))
